/******************************************************************************\

          This file is part of the C! library.  A.K.A the cbang library.

              Copyright (c) 2003-2012, Cauldron Development LLC
                 Copyright (c) 2003-2012, Stanford University
                             All rights reserved.

        The C! library is free software: you can redistribute it and/or
        modify it under the terms of the GNU Lesser General Public License
        as published by the Free Software Foundation, either version 2.1 of
        the License, or (at your option) any later version.

        The C! library is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
        Lesser General Public License for more details.

        You should have received a copy of the GNU Lesser General Public
        License along with the C! library.  If not, see
        <http://www.gnu.org/licenses/>.

        In addition, BSD licensing may be granted on a case by case basis
        by written permission from at least one of the copyright holders.
        You may request written permission by emailing the authors.

                For information regarding this software email:
                               Joseph Coffland
                        joseph@cauldrondevelopment.com

\******************************************************************************/

#include "Message.h"

#include <cbang/String.h>
#include <cbang/log/Logger.h>

#include <sstream>

#include <string.h>

using namespace std;
using namespace cb;
using namespace cb::HTTP;

const char *Message::TIME_FORMAT = "%a, %d %b %Y %H:%M:%S %Z";


bool Message::hasCookie(const string &name) const {
  return cookies.find(name) != cookies.end();
}


const SmartPointer<Cookie> &Message::getCookie(const string &name) const {
  cookies_t::const_iterator it = cookies.find(name);
  if (it == cookies.end()) THROWS("Cookie '" << name << "' not found");
  return it->second;
}


void Message::setCookie(const SmartPointer<Cookie> &cookie) {
  pair<cookies_t::iterator, bool> result =
    cookies.insert(cookies_t::value_type(cookie->getName(), cookie));
  if (!result.second) result.first->second = cookie;
}


void Message::setPersistent(bool x) {
  if (getVersion() < 1.1) {
    if (x) set("Connection", "Keep-Alive");
    else erase("Connection");

  } else if (x) erase("Connection");
  else set("Connection", "close");
}


bool Message::isPersistent() const {
  string connection = String::toLower(get("Connection", ""));
  if (getVersion() < 1.1) return connection == "keep-alive";
  return connection != "close";
}


string Message::getHeader() const {
  ostringstream str;
  str << *this;
  return str.str();
}


void Message::setSessionID(const string &id, const string &name) {
  cookies_t::const_iterator it = cookies.find(name);
  if (it == cookies.end())
    cookies.insert(cookies_t::value_type
                   (name, new Cookie(name, id, "", "/")));
  else it->second->setValue(id);
}


string Message::getSessionID(const string &name) const {
  cookies_t::const_iterator it = cookies.find(name);
  if (it != cookies.end())
    return it->second->getValue();

  if (!has("Cookie")) return "";

  SmartPointer<Cookie> cookie = new Cookie(name);
  cookie->read(get("Cookie"));
  const_cast<Message *>(this)->cookies[name] = cookie;

  return cookie->getValue();
}


string Message::getVersionString() const {
  return String::printf("HTTP/%.01f", version);
}


void Message::readVersionString(const string &s) {
  if (s.length() < 5 || s.substr(0, 5) != "HTTP/")
    THROWS("Missing 'HTTP/' in HTTP-Version '" << s << "'");

  setVersion(String::parseDouble(s.substr(5)));
}


void Message::read(const string &data) {
  istringstream stream(data);
  read(stream);
}


istream &Message::read(istream &stream) {
  char line[1024];
  memset(line, 0, sizeof(line));

  stream.getline(line, sizeof(line));
  if (stream.fail()) THROW("Failed to read stream");
  LOG_DEBUG(6, "Read " << stream.gcount() << " bytes of HTTP header line");

  // Remove trailing space
  int len = strlen(line);
  for (int i = len - 1; 0 <= i && isspace(line[i]); i--)
    line[i] = 0;

  LOG_DEBUG(5, "HTTP header line '" << line << "' length " << len);

  readHeaderLine(line);
  return Header::read(stream);
}


ostream &Message::write(ostream &stream) const {
  stream << getHeaderLine() << "\r\n";

  Header::write(stream);

  // Write cookies
  cookies_t::const_iterator it;
  for (it = cookies.begin(); it != cookies.end(); it++)
    it->second->write(stream, isResponse());

  return stream << "\r\n";
}
