/******************************************************************************\

          This file is part of the C! library.  A.K.A the cbang library.

              Copyright (c) 2003-2012, Cauldron Development LLC
                 Copyright (c) 2003-2012, Stanford University
                             All rights reserved.

        The C! library is free software: you can redistribute it and/or
        modify it under the terms of the GNU Lesser General Public License
        as published by the Free Software Foundation, either version 2.1 of
        the License, or (at your option) any later version.

        The C! library is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
        Lesser General Public License for more details.

        You should have received a copy of the GNU Lesser General Public
        License along with the C! library.  If not, see
        <http://www.gnu.org/licenses/>.

        In addition, BSD licensing may be granted on a case by case basis
        by written permission from at least one of the copyright holders.
        You may request written permission by emailing the authors.

                For information regarding this software email:
                               Joseph Coffland
                        joseph@cauldrondevelopment.com

\******************************************************************************/

#include "WebServer.h"

#include <cbang/SStream.h>
#include <cbang/util/SmartLock.h>
#include <cbang/time/Time.h>
#include <cbang/time/Timer.h>
#include <cbang/security/MD5.h>
#include <cbang/config/Options.h>

using namespace std;
using namespace cb;
using namespace cb::HTTP;


WebServer::WebServer(Options &options, const string &match,
                     Script::Handler *parent, hasFeature_t hasFeature) :
  Server(options, hasFeature(FEATURE_SSL) ? &sslCtx : 0),
  WebHandler(options, match, parent, hasFeature),
  sessionTimeout(Time::SEC_PER_HOUR), lastSessionCleanup(0) {
  
  options.pushCategory("Web Server");

  if (hasFeature(FEATURE_SESSIONS))
    options.addTarget("session-timeout", sessionTimeout, "Time since last use "
                      "before session IDs are dropped.");

  options.popCategory();
}


bool WebServer::_hasFeature(int feature) {
  switch (feature) {
  case FEATURE_SESSIONS: return true;
  case FEATURE_SSL: return false;

  default: return WebHandler::_hasFeature(feature);
  }
}


void WebServer::init() {
  Server::addHandler(this);
  Server::init();
  WebHandler::init();
}


SessionPtr WebServer::openSession(WebContext &ctx, const string &_id) {
  if (!hasFeature(FEATURE_SESSIONS)) return 0;

  string id = _id;
  if (id.empty()) {
    MD5 md5;
    uint128_t idNum;

    if (ctx.getRequest().has("User-Agent"))
      md5.add(ctx.getRequest().get("User-Agent"));
    md5.add(Time::now());
    md5.getDigest((unsigned char *)&idNum);

    id = SSTR(idNum);
  }

  SmartLock lock(this);

  // Get the session
  SessionPtr session;
  sessions_t::iterator it = sessions.find(id);
  if (it == sessions.end()) {
    session = createSession(id);
    sessions.insert(sessions_t::value_type(id, session));
    return session;
  }

  it->second->touch(); // Update timestamp
  return it->second;
}


void WebServer::buildResponse(HTTP::Context *_ctx) {
  init();

  // Cleanup old sessions
  if (hasFeature(FEATURE_SESSIONS)) {
    SmartLock lock(this);

    uint64_t now = Time::now();
    if (lastSessionCleanup + 5 < now) {
      lastSessionCleanup = now;

      vector<string> remove;
      sessions_t::iterator it;
      for (it = sessions.begin(); it != sessions.end(); it++)
        if (it->second->getLastUsed() + sessionTimeout < now)
          remove.push_back(it->first);

      for (unsigned i = 0; i < remove.size(); i++)
        sessions.erase(remove[i]);
    }
  }

  WebHandler::buildResponse(_ctx);
}
