/******************************************************************************\

          This file is part of the C! library.  A.K.A the cbang library.

              Copyright (c) 2003-2012, Cauldron Development LLC
                 Copyright (c) 2003-2012, Stanford University
                             All rights reserved.

        The C! library is free software: you can redistribute it and/or
        modify it under the terms of the GNU Lesser General Public License
        as published by the Free Software Foundation, either version 2.1 of
        the License, or (at your option) any later version.

        The C! library is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
        Lesser General Public License for more details.

        You should have received a copy of the GNU Lesser General Public
        License along with the C! library.  If not, see
        <http://www.gnu.org/licenses/>.

        In addition, BSD licensing may be granted on a case by case basis
        by written permission from at least one of the copyright holders.
        You may request written permission by emailing the authors.

                For information regarding this software email:
                               Joseph Coffland
                        joseph@cauldrondevelopment.com

\******************************************************************************/

#include "Reader.h"

#include "None.h"
#include "Boolean.h"
#include "Number.h"
#include "String.h"
#include "List.h"
#include "Dict.h"

#include <cbang/String.h>

#include <vector>
#include <cctype>

using namespace std;
using namespace cb::PyON;


ValuePtr Reader::parse() {
  ValuePtr value;

  while (good()) {
    switch (next()) {
    case 'N': case 'n':
      value = new None;
      value->read(*this);
      return value;

    case 'T': case 't': case 'F': case 'f':
      value = new Boolean;
      value->read(*this);
      return value;

    case '-': case '.':
    case '0': case '1': case '2': case '3': case '4':
    case '5': case '6': case '7': case '8': case '9':
      value = new Number;
      value->read(*this);
      return value;

    case '"':
      value = new String;
      value->read(*this);
      return value;

    case '[':
      value = new List;
      value->read(*this);
      return value;

    case '{':
      value = new Dict;
      value->read(*this);
      return value;

    default: match("NnTtFf-.0123456789\"[{");
    }
  }

  THROWS("Parse failed at " << line << ':' << column);
}


char Reader::get() {
  char c = stream.get();

  if (c == '\n') {
    line++;
    column = 0;

  } else if (c != '\r') column++;

  return c;
}


char Reader::next() {
  while (good())
    switch (stream.peek()) {
    case '\n': case '\r': case '\t': case ' ': get(); break;

    case '#':
      while (stream.peek() != '\n') get();
      break;

    default: return stream.peek();
    }

  THROWS("Parse failed at " << line << ':' << column);
}


char Reader::match(const char *chars) {
  char x = next();

  for (int i = 0; chars[i]; i++)
    if (x == chars[i]) {
      advance();
      return x;
    }

  THROWS("Expected one of '" << cb::String::escapeC(chars)
         << "' but found '" << cb::String::escapeC(string(1, x))
         << "' at " << line << ':' << column);
}


const string Reader::parseKeyword() {
  string s;
  while (good() && isalpha(stream.peek())) s += stream.get();
  return s;
}


bool Reader::parseBoolean() {
  string value = cb::String::toLower(parseKeyword());

  if (value == "true") return true;
  else if (value == "false") return false;
  else THROWS("Expected keyword 'true' or 'false' but found '" << value
              << "' at " << line << ':' << column);
}


double Reader::parseNumber() {
  string value;
  bool first = true;

  while (good()) {
    char c = next();

    switch (c) {
    case '-': if (!first) break;
    case '.':
    case '0': case '1': case '2': case '3': case '4':
    case '5': case '6': case '7': case '8': case '9':
      value += c;
      advance();
      first = false;
      continue;
    }

    break;
  }

  return cb::String::parseDouble(value);
}


const string Reader::parseString() {
  match("\"");

  string s;
  bool escape = false;
  while (good()) {
    char c = get();

    if (c == '"' && !escape) break;
    if (good()) s += c;

    escape = c == '\\';
  }

  return cb::String::unescapeC(s);
}
