/******************************************************************************\

          This file is part of the C! library.  A.K.A the cbang library.

              Copyright (c) 2003-2012, Cauldron Development LLC
                 Copyright (c) 2003-2012, Stanford University
                             All rights reserved.

        The C! library is free software: you can redistribute it and/or
        modify it under the terms of the GNU Lesser General Public License
        as published by the Free Software Foundation, either version 2.1 of
        the License, or (at your option) any later version.

        The C! library is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
        Lesser General Public License for more details.

        You should have received a copy of the GNU Lesser General Public
        License along with the C! library.  If not, see
        <http://www.gnu.org/licenses/>.

        In addition, BSD licensing may be granted on a case by case basis
        by written permission from at least one of the copyright holders.
        You may request written permission by emailing the authors.

                For information regarding this software email:
                               Joseph Coffland
                        joseph@cauldrondevelopment.com

\******************************************************************************/

#include "Cipher.h"

#include <cbang/Zap.h>
#include <cbang/Exception.h>

#include <openssl/evp.h>
#include <openssl/err.h>

#include <string.h>

using namespace std;
using namespace cb;


Cipher::Cipher(const EVP_CIPHER *type, bool encrypt, const void *key,
               const void *iv, ENGINE *impl) :
  ctx(new EVP_CIPHER_CTX), encrypt(encrypt) {
  EVP_CIPHER_CTX_init(ctx);

  if (encrypt) {
    if (!EVP_EncryptInit_ex(ctx, type, impl, (const unsigned char *)key,
                            (const unsigned char *)iv))
      THROWS("Cipher encrypt init failed: " << getErrorStr());

  } else {
    if (!EVP_DecryptInit_ex(ctx, type, impl, (const unsigned char *)key,
                            (const unsigned char *)iv))
      THROWS("Cipher decrypt init failed: " << getErrorStr());
  }
}


Cipher::~Cipher() {
  EVP_CIPHER_CTX_cleanup(ctx);
  zap(ctx);
}


unsigned Cipher::blockSize() const {
  return EVP_CIPHER_CTX_block_size(ctx);
}


unsigned Cipher::update(void *out, unsigned length, const void *in,
                        unsigned inLen) {
  SmartPointer<char>::Array buf;
  if (out == in) {
    if (inLen < length)
      THROW("If using cipher in place then the intput buffer length must be "
            "at least as large as the output buffer");

    buf = new char[inLen];
    memcpy(buf.get(), in, inLen);
    in = buf.get();
  }

  if (encrypt) {
    if (!EVP_EncryptUpdate(ctx, (unsigned char *)out, (int *)&length,
                           (unsigned char *)in, inLen))
      THROWS("Cipher encrypt update failed: " << getErrorStr());

  } else {
    if (!EVP_DecryptUpdate(ctx, (unsigned char *)out, (int *)&length,
                           (unsigned char *)in, inLen))
      THROWS("Cipher decrypt update failed: " << getErrorStr());
  }

  return length;
}


unsigned Cipher::final(void *out, unsigned length) {
  if (encrypt) {
    if (!EVP_EncryptFinal_ex(ctx, (unsigned char *)out, (int *)&length))
      THROWS("Cipher encrypt final failed: " << getErrorStr());

  } else {
    if (!EVP_DecryptFinal_ex(ctx, (unsigned char *)out, (int *)&length))
      THROWS("Cipher decrypt final failed: " << getErrorStr());
  }

  return length;
}


string Cipher::getErrorStr() {
  ERR_load_crypto_strings();
  return ERR_error_string(ERR_get_error(), 0);
}
