/******************************************************************************\

          This file is part of the C! library.  A.K.A the cbang library.

              Copyright (c) 2003-2012, Cauldron Development LLC
                 Copyright (c) 2003-2012, Stanford University
                             All rights reserved.

        The C! library is free software: you can redistribute it and/or
        modify it under the terms of the GNU Lesser General Public License
        as published by the Free Software Foundation, either version 2.1 of
        the License, or (at your option) any later version.

        The C! library is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
        Lesser General Public License for more details.

        You should have received a copy of the GNU Lesser General Public
        License along with the C! library.  If not, see
        <http://www.gnu.org/licenses/>.

        In addition, BSD licensing may be granted on a case by case basis
        by written permission from at least one of the copyright holders.
        You may request written permission by emailing the authors.

                For information regarding this software email:
                               Joseph Coffland
                        joseph@cauldrondevelopment.com

\******************************************************************************/

#define NOCRYPT // For WIN32 to avoid conflicts with openssl

#include "KeyPair.h"

#include <cbang/Exception.h>
#include <cbang/SmartPointer.h>
#include <cbang/log/Logger.h>

#include <iostream>

#include "BIStream.h"
#include "BOStream.h"

#include <openssl/pem.h>
#include <openssl/dsa.h>
#include <openssl/engine.h>

using namespace cb;
using namespace std;

static void gen_callback(int cmd, int x, ostream *stream) {
  if (cmd > 0) *stream << '.' << flush;
}


KeyPair::~KeyPair() {
  if (key) DSA_free(key);
}


void KeyPair::generate() {
  if (key) DSA_free(key);

  Logger::LogStream stream = LOG_INFO_STREAM(1);

  *stream << "Generating key pair:" << flush;

  key = DSA_generate_parameters(1024, 0, 0, 0, 0,
                                (void (*)(int, int, void *))gen_callback,
                                stream.get());

  if (!key) THROW("Failed to generate key pair parameters");

  if (!DSA_generate_key(key))
    THROW("Failed to generate new key pair");
}


bool KeyPair::hasPublic() const {
  return key && key->pub_key;
}


bool KeyPair::hasPrivate() const {
  return key && key->priv_key;
}


const string KeyPair::sign(const string &s) const {
  unsigned size = DSA_size(key);
  SmartPointer<unsigned char>::Array buf = new unsigned char [size];

  if (!DSA_sign(0, (unsigned char *)s.c_str(), s.length(), buf.get(), &size,
                key))
    THROW("Failed to sign message");

  return string((char *)buf.get(), size);
}


istream &KeyPair::read(istream &stream) {
  BIStream bio(stream);

  if (key) DSA_free(key);

  key = PEM_read_bio_DSA_PUBKEY(bio.getBIO(), 0, 0, 0);
  if (!key) THROW("Failed to read key");

  PEM_read_bio_DSAPrivateKey(bio.getBIO(), &key, 0, 0);

  return stream;
}


ostream &KeyPair::write(ostream &stream) const {
  BOStream bio(stream);

  if (!key) THROW("No key loaded");

  if (hasPublic())
    if (!PEM_write_bio_DSA_PUBKEY(bio.getBIO(), key))
      THROW("Failed to write public key");

  if (hasPrivate())
    if (!PEM_write_bio_DSAPrivateKey(bio.getBIO(), key, 0, 0, 0, 0, 0))
      THROW("Failed to write private key");

  return stream;
}
