/******************************************************************************\

          This file is part of the C! library.  A.K.A the cbang library.

              Copyright (c) 2003-2012, Cauldron Development LLC
                 Copyright (c) 2003-2012, Stanford University
                             All rights reserved.

        The C! library is free software: you can redistribute it and/or
        modify it under the terms of the GNU Lesser General Public License
        as published by the Free Software Foundation, either version 2.1 of
        the License, or (at your option) any later version.

        The C! library is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
        Lesser General Public License for more details.

        You should have received a copy of the GNU Lesser General Public
        License along with the C! library.  If not, see
        <http://www.gnu.org/licenses/>.

        In addition, BSD licensing may be granted on a case by case basis
        by written permission from at least one of the copyright holders.
        You may request written permission by emailing the authors.

                For information regarding this software email:
                               Joseph Coffland
                        joseph@cauldrondevelopment.com

\******************************************************************************/

#include "RSAKey.h"

#include "SSL.h"
#include "BOStream.h"

#include <cbang/Exception.h>
#include <cbang/SmartPointer.h>

// This avoids a conflict with OCSP_RESPONSE in wincrypt.h
#ifdef OCSP_RESPONSE
#undef OCSP_RESPONSE
#endif

#include <openssl/rsa.h>
#include <openssl/pem.h>

#include <string.h>
#include <stdlib.h>


using namespace cb;
using namespace std;


namespace {
  void sp_free_bio(void *bio) {
    BIO_free((BIO *)bio);
  }
}


RSAKey::RSAKey(const string &pem, bool pub) : key(0), pub(pub) {
  SSL::init();

  SmartPointer<BIO, RefCounter<BIO, DeallocFunc<sp_free_bio> > >
    pemBIO = BIO_new_mem_buf((void *)pem.c_str(), pem.length());
  
  // Load key
  if (pub) {
    key = PEM_read_bio_RSA_PUBKEY(pemBIO.get(), 0, 0, 0);
      
    // Try PKCS1 format
    if (!key) {
      (void)BIO_reset(pemBIO.get());
      key = PEM_read_bio_RSAPublicKey(pemBIO.get(), 0, 0, 0);
    }
    
  } else key = PEM_read_bio_RSAPrivateKey(pemBIO.get(), 0, 0, 0);
  
  if (!key) THROWS("Error loading RSA key: " << cb::SSL::getErrorStr());
}


RSAKey::~RSAKey() {
  release();
}


unsigned RSAKey::size() const {
  return key ? RSA_size(key) : 0;
}


static void gen_callback(int x, int n, void *data) {
  if (data) (*(KeyGenCallback *)data)(x, n);
}


void RSAKey::generate(unsigned size, unsigned long e,
                      SmartPointer<KeyGenCallback> callback) {
  release();
  key = RSA_generate_key(size, e, gen_callback, callback.get());
  if (!key) THROWS("Error generating RSA pair: " << cb::SSL::getErrorStr());
  pub = false;
}


static int convertPadding(int padding) {
  switch (padding) {
  case RSAKey::NO_PADDING: return RSA_NO_PADDING;
  case RSAKey::PKCS1_PADDING: return RSA_PKCS1_PADDING;
  case RSAKey::SSLV23_PADDING: return RSA_SSLV23_PADDING;
  case RSAKey::PKCS1_OAEP_PADDING: return RSA_PKCS1_OAEP_PADDING;
  case RSAKey::X931_PADDING: return RSA_X931_PADDING;
  }

  return padding;
}


unsigned RSAKey::encrypt(const void *from, void *to, int padding) const {
  if (!key) THROW("Invalid RSA key");

  SmartPointer<char>::Array tmp;
  if (from == to) {
    tmp = new char[size()];
    memcpy(tmp.get(), from, size());
    from = tmp.get();
  }

  padding = convertPadding(padding);

  int ret;
  if (pub) ret = RSA_public_encrypt(size(), (unsigned char *)from,
                                    (unsigned char *)to, key, padding);
  else ret = RSA_private_encrypt(size(), (unsigned char *)from,
                                 (unsigned char *)to, key, padding);

  if (ret < 0) THROWS("RSA encrypt failed: " << cb::SSL::getErrorStr());

  return (unsigned)ret;
}


unsigned RSAKey::decrypt(const void *from, void *to, int padding) const {
  if (!key) THROW("Invalid RSA key");

  SmartPointer<char>::Array tmp;
  if (from == to) {
    tmp = new char[size()];
    memcpy(tmp.get(), from, size());
    from = tmp.get();
  }

  padding = convertPadding(padding);

  int ret;
  if (pub) ret = RSA_public_decrypt(size(), (unsigned char *)from,
                                    (unsigned char *)to, key, padding);
  else ret = RSA_private_decrypt(size(), (unsigned char *)from,
                                 (unsigned char *)to, key, padding);

  if (ret < 0) THROWS("RSA encrypt failed: " << cb::SSL::getErrorStr());

  return (unsigned)ret;
}


void RSAKey::release() {
  if (key) {
    RSA_free(key);
    key = 0;
  }
}


ostream &RSAKey::print(ostream &stream) const {
  if (!key) THROW("Invalid RSA key");

  BOStream bio(stream);
  if (pub) PEM_write_bio_RSAPublicKey(bio.getBIO(), key);
  else PEM_write_bio_RSAPrivateKey(bio.getBIO(), key, 0, 0, 0, 0, 0);

  return stream;
}
