/******************************************************************************\

    OpenSCAM is an Open-Source CAM software.
    Copyright (C) 2011-2012 Joseph Coffland <joseph@cauldrondevelopment.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

\******************************************************************************/

#include "Project.h"

#include <openscam/sim/ToolTable.h>

#include <cbang/os/SystemUtilities.h>
#include <cbang/time/Time.h>
#include <cbang/log/Logger.h>
#include <cbang/xml/XMLWriter.h>
#include <cbang/xml/XMLReader.h>

using namespace std;
using namespace cb;
using namespace OpenSCAM;


Project::Project(Options &_options, ToolTable &toolTable,
                 const std::string &filename) :
  options(_options), resolutionMode(ResolutionMode::RESOLUTION_MEDIUM),
  resolution(1), toolTable(toolTable), workpieceMargin(5), watch(true),
  lastWatch(0) {

  options.setAllowReset(true);

  options.pushCategory("Renderer");
  options.addTarget("resolution-mode", resolutionMode, "Automatically compute "
                    "a reasonable renderer grid resolution.  Valid values "
                    "are 'low', 'medium', 'high', 'manual'.  If 'manual' then "
                    "'resolution' will be used.");
  options.addTarget("resolution", resolution, "Renderer grid resolution");
  options.popCategory();

  options.pushCategory("NC Files");
  options.addTarget("watch", watch, "Watch input files for changes and "
                    "automatically reload");
  options.add("nc-files", "GCode files")->setType(Option::STRINGS_TYPE);
  options.popCategory();

  options.pushCategory("Workpiece");
  options.add("automatic-workpiece", "Automatically compute a cuboid "
              "workpiece based on the tool path boundary")->setDefault(true);
  options.addTarget("workpiece-margin", workpieceMargin,
                    "Percent margin around automatic workpiece");
  options.addTarget("workpiece-min", workpieceMin,
                    "Minimum bound of cuboid workpiece");
  options.addTarget("workpiece-max", workpieceMax,
                    "Maximum bound of cuboid workpiece");
  options.popCategory();

  if (!filename.empty()) load(filename);
}


void Project::setFilename(const string &_filename) {
  if (_filename.empty() || filename == _filename) return;
  filename = _filename;

  vector<string> absFiles = getAbsoluteFiles();

  Option &option = options["nc-files"];
  option.reset();

  for (unsigned i = 0; i < absFiles.size(); i++)
    option.append(makeRelative(absFiles[i]));
}


string Project::makeRelative(const string &path) const {
  string dir = filename.empty() ?
    SystemUtilities::getcwd() : SystemUtilities::dirname(filename);
  return SystemUtilities::relative(dir, path, 4);
}


string Project::makeAbsolute(const string &path) const {
  string dir = filename.empty() ?
    SystemUtilities::getcwd() : SystemUtilities::dirname(filename);
  return SystemUtilities::makeRelative(dir, path);
}


void Project::load(const string &_filename) {
  setFilename(_filename);

  XMLReader reader;
  reader.addFactory("tool_table", &toolTable);
  reader.read(filename, &options);

  // Load NC files
  Option::strings_t ncFiles = options["nc-files"].toStrings();
  options["nc-files"].reset();
  for (unsigned i = 0; i < ncFiles.size(); i++) addFile(ncFiles[i]);
}


void Project::save(const string &_filename) {
  setFilename(_filename);

  SmartPointer<iostream> stream = SystemUtilities::open(filename, ios::out);
  XMLWriter writer(*stream, true);

  options["resolution-mode"].set(resolutionMode.toString());

  writer.startElement("openscam");
  options.write(writer, 0);
  toolTable.write(writer);
  writer.endElement("openscam");
}


void Project::addFile(const string &path) {
  string abs = makeAbsolute(path);
  if (files.has(abs)) return; // Duplicate
  files[abs] = SystemUtilities::getModificationTime(abs);
  options["nc-files"].append(makeRelative(abs));
}


vector<string> Project::getRelativeFiles() const {
  return options["nc-files"].toStrings();
}


vector<string> Project::getAbsoluteFiles() const {
  vector<string> absFiles;

  for (files_t::const_iterator it = files.begin(); it != files.end(); it++)
    absFiles.push_back(it->first);

  return absFiles;
}


void Project::setFiles(vector<string> &files) {
  options["nc-files"].reset();
  this->files.clear();
  for (unsigned i = 0; i < files.size(); i++) addFile(files[i]);
}


bool Project::getAutomaticWorkpiece() const {
  return options["automatic-workpiece"].toBoolean() ||
    workpieceMin.empty() || workpieceMax.empty();
}


void Project::setWorkpieceBounds(const Rectangle3D &bounds) {
  options["workpiece-min"].set(bounds.getMin().toString());
  options["workpiece-max"].set(bounds.getMax().toString());
}


Rectangle3D Project::getWorkpieceBounds() const {
  Vector3D wpMin = workpieceMin.empty() ? Vector3D() : Vector3D(workpieceMin);
  Vector3D wpMax = workpieceMax.empty() ? Vector3D() : Vector3D(workpieceMax);
  return Rectangle3D(wpMin, wpMax);
}


bool Project::checkFiles() {
  if (watch && lastWatch < Time::now()) {
    for (files_t::iterator it = files.begin(); it != files.end(); it++) {
      uint64_t fileTime = SystemUtilities::getModificationTime(it->first);
      if (it->second < fileTime) {
        it->second = fileTime;
        LOG_INFO(1, "File changed: " << it->first);
        return true;
      }
    }

    lastWatch = Time::now();
  }

  return false;
}
