/******************************************************************************\

    OpenSCAM is an Open-Source CAM software.
    Copyright (C) 2011-2012 Joseph Coffland <joseph@cauldrondevelopment.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

\******************************************************************************/

#include "SimulationManager.h"

#include "Workpiece.h"
#include "ToolPath.h"
#include "CutWorkpiece.h"
#include "Project.h"
#include "Sweep.h"

#include <limits>

using namespace std;
using namespace cb;
using namespace OpenSCAM;


SimulationManager::SimulationManager(Options &options) :
  controller(options, *this) {}


// Keep member destructors out of header
SimulationManager::~SimulationManager() {}


void SimulationManager::init() {
  controller.init();
}


void SimulationManager::reset() {
  controller.reset();
  path = 0;
  cutWP = 0;
}


void SimulationManager::load(Project &project) {
  // Load G code
  controller.reset();
  Machine::start();
  const ToolTable &tools = controller.getToolTable();
  path = new ToolPath(tools);

  // Reload files
  vector<string> files = project.getAbsoluteFiles();
  for (unsigned i = 0; i < files.size(); i++)
    Interpreter(controller).read(files[i]);

  // Workpiece bounds
  Rectangle3R wpBounds;
  if (!project.getAutomaticWorkpiece()) wpBounds = project.getWorkpieceBounds();
  else {
    // Guess workpiece bounds from cutting moves
    real lowestZ = numeric_limits<real>::max();

    vector<SmartPointer<Sweep> > sweeps;
    vector<Rectangle3R> bboxes;
    bool first = true;
    for (unsigned i = 0; i < path->size(); i++) {
      const Move &move = path->at(i);
      if (move.getType() != MoveType::MOVE_RAPID) {
        if (first) {
          lowestZ = move.getStartPt().z();
          first = false;
        }

        unsigned tool = move.getTool();
        if (sweeps.size() <= tool) sweeps.resize(tool + 1);
        if (sweeps[tool].isNull()) sweeps[tool] = tools.get(tool)->getSweep();

        sweeps[tool]->getBBoxes(move.getStartPt(), move.getEndPt(), bboxes, 0);
      }
    }

    for (unsigned i = 0; i < bboxes.size(); i++) wpBounds.add(bboxes[i]);

    // Start from lowestZ
    if (lowestZ != numeric_limits<real>::max()) {
      Vector3R bMin = wpBounds.getMin();
      Vector3R bMax = wpBounds.getMax();
      wpBounds = Rectangle3R(bMin, Vector3D(bMax.x(), bMax.y(), lowestZ));

      // At least a height of 2
      if (wpBounds.getHeight() < 2)
        wpBounds =
          Rectangle3R(bMin, Vector3D(bMax.x(), bMax.y(), bMin.z() + 2));
    }

    if (wpBounds != Rectangle3R()) {
      // Margin
      Vector3R margin =
        wpBounds.getDimensions() * project.getWorkpieceMargin() / 100.0;
      wpBounds.add(wpBounds.getMin() - margin);
      wpBounds.add(wpBounds.getMax() + Vector3R(margin.x(), margin.y(), 0));

      project.setWorkpieceBounds(wpBounds);
    }
  }

  // Auto resolution
  if (wpBounds != Rectangle3R() &&
      project.getResolutionMode() != ResolutionMode::RESOLUTION_MANUAL) {
    double divisor;
    switch (project.getResolutionMode()) {
    case ResolutionMode::RESOLUTION_LOW: divisor = 100000; break;
    case ResolutionMode::RESOLUTION_HIGH: divisor = 5000000; break;
    case ResolutionMode::RESOLUTION_VERY_HIGH: divisor = 10000000; break;
    default: divisor = 250000; break; // Medium
    }

    project.setResolution(pow(wpBounds.getVolume() / divisor, 1.0 / 3.0));
  }

  // Setup cut simulation
  SmartPointer<Workpiece> wp = new Workpiece(wpBounds);
  SmartPointer<ToolSweep> ts = new ToolSweep(path);
  cutWP = new CutWorkpiece(ts, wp);
}


void SimulationManager::move(const Move &move) {
  Machine::move(move);
  path->add(move);
}
