/******************************************************************************\

    OpenSCAM is an Open-Source CAM software.
    Copyright (C) 2011-2012 Joseph Coffland <joseph@cauldrondevelopment.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

\******************************************************************************/

#include "Filter.h"

#include <openscam/gcode/Parser.h>

#include <cbang/String.h>

#include <cbang/config/Options.h>

#include <cctype>

using namespace std;
using namespace cb;
using namespace OpenSCAM;


Filter::Filter(Options &options, std::ostream &stream,
               const std::string &f) :
  Printer(stream), options(options), controller(options, *this),
  interp(controller), f(f) {

  options.pushCategory("Filter");
  options.addTarget("filter", this->f, "Filter of the form "
                    "<letter><operator><value> where <operator> is one of "
                    "==, != <, >, <=, >=");
  options.popCategory();
}


void Filter::init() {
  if (!options["filter"].hasValue()) THROW("filter not set");

  string f = options["filter"];
  if (!f.empty()) {
    if (f.size() < 3) THROWS("Invalid filter '" << f << "'");
    
    // letter
    if (!isalpha(f[0])) THROWS("Filter must start with letter");
    var = toupper(f[0]);
    
    // Operator
    if (f.substr(1, 2) == "==") op = "==";
    else if (f.substr(1, 2) == "!=") op = "!=";
    else if (f.substr(1, 2) == "<=") op = "<=";
    else if (f.substr(1, 2) == ">=") op = ">=";
    else if (f.substr(1, 1) == "<")  op = "<";
    else if (f.substr(1, 1) == ">")  op = ">";
    else THROW("Invalid filter operator");
    
    // Value
    value = String::parseDouble(f.substr(op.length() + 1));
  }
}


void Filter::read(const InputSource &source) {
  try {
    Parser().parse(source, *this);
  } catch (const EndProgram &) {}
}


void Filter::move(const Move &move) {
  if (move.getType() == MoveType::MOVE_CUTTING) {
    Segment3D seg = move;
    if (seg.findSmallest()) seg.reverse();
    if (moves.insert(seg).second) uniqueMove = true;
    moved = true;
  }
}


void Filter::operator()(const SmartPointer<Block> &block) {
  moved = uniqueMove = false;

  interp(block);

  if (moved && !uniqueMove) return;

  double varValue = controller.getVar(var);

  if (!op.empty()) {
    if (op == "==" && varValue != value) return;
    if (op == "!=" && varValue == value) return;
    if (op == "<"  && varValue >= value) return;
    if (op == ">"  && varValue <= value) return;
    if (op == "<=" && varValue > value) return;
    if (op == ">=" && varValue < value) return;
  }

  Printer::operator()(block);
}
