/******************************************************************************\

    OpenSCAM is an Open-Source CAM software.
    Copyright (C) 2011-2012 Joseph Coffland <joseph@cauldrondevelopment.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

\******************************************************************************/

#include "GUIApp.h"

#include "GtkObserver.h"
#include "GtkRangeObserver.h"
#include "GtkObjectSet.h"

#include <openscam/Geom.h>
#include <openscam/view/GL.h>
#include <openscam/view/Viewer.h>
#include <openscam/cutsim/SimulationManager.h>
#include <openscam/cutsim/Project.h>
#include <openscam/remote/ConnectionManager.h>
#include <openscam/render/Renderer.h>
#include <openscam/stl/STL.h>

#include <cbang/Exception.h>
#include <cbang/SmartPointer.h>
#include <cbang/String.h>
#include <cbang/log/Logger.h>
#include <cbang/util/DefaultCatch.h>
#include <cbang/util/Resource.h>
#include <cbang/os/SystemUtilities.h>
#include <cbang/time/TimeInterval.h>

#include <gtk/gtkgl.h>

using namespace std;
using namespace cb;
using namespace OpenSCAM;

namespace OpenSCAM {
  extern const DirectoryResource resource0;
#include "Images.dat"
}


GTK_SIGNAL1V(on_render_mode_changed, GtkComboBox *, GUIApp);

GTK_SIGNAL1V(on_program_scrollbar_changed, GtkRange *, GUIApp);

GTK_SIGNAL1V(on_check_button_toggled, GtkToggleButton *, GUIApp);
GTK_SIGNAL1V(on_tool_button_clicked, GtkWidget *, GUIApp);

GTK_SIGNAL1V(on_menu_item_activate, GtkMenuItem *, GUIApp);

GTK_SIGNAL2B(on_glarea_mouse_press, GtkWidget *, GdkEventButton *, GUIApp);
GTK_SIGNAL2B(on_glarea_mouse_motion, GtkWidget *, GdkEventMotion *, GUIApp);
GTK_SIGNAL2B(on_glarea_mouse_scroll, GtkWidget *, GdkEventScroll *, GUIApp);

GTK_SIGNAL1V(on_glarea_realize, GtkWidget *, GUIApp);
GTK_SIGNAL2B(on_glarea_configure, GtkWidget *, GdkEventConfigure *, GUIApp);
GTK_SIGNAL2B(on_glarea_expose, GtkWidget *, GdkEventExpose *, GUIApp);

GTK_SIGNAL1V(on_window_destroy, GtkWindow *, GUIApp);
GTK_SIGNAL2B(on_window_delete_event, GtkWindow *, GdkEvent *, GUIApp);

GTK_SIGNAL0B(on_idle, GUIApp);
GTK_SIGNAL2B(on_map_event, GtkWidget *, GdkEvent *, GUIApp);
GTK_SIGNAL2B(on_unmap_event, GtkWidget *, GdkEvent *, GUIApp);


GUIApp::GUIApp() :
  Application("OpenSCAM"), simulationManager(new SimulationManager(options)),
  connectionManager(new ConnectionManager(options)),
  renderer(new Renderer(options)), view(new View(options, valueSet, renderer)),
  viewer(new Viewer), dirty(true), idleID(0), window(0), glArea(0),
  playButton(0) {

  // Configure commandline
  cmdLine.setAllowConfigAsFirstArg(false);
  cmdLine.setAllowPositionalArgs(true);

  view->clearFlag(View::SHOW_INFO_FLAG);
}


GUIApp::~GUIApp() {
  if (!renderer.isNull()) renderer->join();
}


void GUIApp::setProjectFile(const string &filename) {
  string projectFile = filename;

  string cwd = SystemUtilities::getcwd();
  if (String::startsWith(projectFile, cwd + '/'))
    projectFile = projectFile.substr(cwd.length() + 1);

  getFileDialog().setFilename(projectFile);

  string title = string("OpenSCAM - ") + projectFile;
  gtk_window_set_title(window, title.c_str());
  gtk_window_set_title(glWindow, title.c_str());
}


void GUIApp::loadImage(const char *name, const guint8 data[]) {
  GdkPixbuf *pixbuf = gdk_pixbuf_new_from_inline(-1, data, false, 0);
  if (!pixbuf) LOG_ERROR("Loading pixbuf " << name);
  GtkImage *image = builder->get<GtkImage>(String::printf("%s_image", name));
  gtk_image_set_from_pixbuf(image, pixbuf);
}


int GUIApp::init(int argc, char *argv[]) {
  // Parse command line, etc
  Application::init(argc, argv);

  // Init
  simulationManager->init();
  connectionManager->init();
  renderer->start();
  view->init();

  // GUI
  gtk_init(&argc, &argv);
  gtk_gl_init(&argc, &argv);

  // Default font
  GtkSettings *settings = gtk_settings_get_default();
  gtk_settings_set_string_property(settings, "gtk-font-name", "Monospace 10",
                                   "");

  // Build GUI
  builder = new Builder(gtk_builder_new());

  const Resource *data = OpenSCAM::resource0.find("openscam.glade");
  if (!data) THROWS("Failed to find glade resource");

  builder->addFromString(data->getData(), data->getLength());

  // Get Widgets
  window = builder->get<GtkWindow>("window");
  glWindow = builder->get<GtkWindow>("glwindow");
  glArea = builder->get<GtkWidget>("glarea");
  playButton = builder->get<GtkToolButton>("play_button");
  aboutDialog = new Dialog(*builder, "about");
  fileDialog = new FileDialog(*this);
  projectDialog = new ProjectDialog(*this);

  // Icons
  loadImage("isometric", isometric_image);
  loadImage("front", front_image);
  loadImage("back", back_image);
  loadImage("left", left_image);
  loadImage("right", right_image);
  loadImage("top", top_image);
  loadImage("bottom", bottom_image);

  // Configure GL area
  GdkGLConfig *glconfig =
    gdk_gl_config_new_by_mode((GdkGLConfigMode)
                              (GDK_GL_MODE_RGB | GDK_GL_MODE_DEPTH |
                               GDK_GL_MODE_DOUBLE));
  if (!glconfig) THROW("No appropriate OpenGL visual found");

  gtk_widget_set_gl_capability(glArea, glconfig, 0, true, GDK_GL_RGBA_TYPE);

  // Connect signals
  builder->connect(this);
  builder->connect(G_OBJECT(glArea), "expose-event",
                   G_CALLBACK(::on_glarea_expose), this);

  // Connect values
  GtkObjectSet valueWidgets(*builder, "_value");
  GtkObjectSet::iterator it;
  for (it = valueWidgets.begin(); it != valueWidgets.end(); it++)
    try {
      valueSet[it->first]->add(new GtkObserver((GtkWidget *)it->second));
    } catch (const Exception &e) {LOG_DEBUG(1, e.getMessage());}

  // Ranges
  GtkRange *programRange = builder->get<GtkRange>("program_scrollbar");
  valueSet["time_ratio"]->add(new GtkRangeObserver(programRange));

  // Play direction
  valueSet["play_direction"]->add(this, &GUIApp::directionUpdate);
  directionButton = builder->get<GtkToolButton>("direction_button");

  valueSet.updated();
  view->path->setByRatio(1);

  // Renderer
  renderProgress = builder->get<GtkProgressBar>("render_progressbar");

  // Show windows
  gtk_window_present(window);
  int x, y, w, h;
  gtk_window_get_position(window, &x, &y);
  gtk_window_get_size(window, &w, &h);

  gtk_window_present(glWindow);
  gtk_window_move(glWindow, x + w + 5, y);

  // Load project
  const vector<string> &args = cmdLine.getPositionalArgs();
  if (1 <= args.size()) {
    openProject(args[0]);

    if (1 < args.size()) LOG_WARNING("Ignoring extra positional arguments");

  } else newProject(false);

  return 0;
}


void GUIApp::run() {
  gtk_main();
}


void GUIApp::rerender() {
  SmartPointer<CutWorkpiece> cutWP = simulationManager->getCutWorkpiece();
  cutWP->getToolSweep()->setTime(view->getTime());
  renderer->setResolution(project->getResolution());
  renderer->reload();
}


void GUIApp::reload() {
  try {
    simulationManager->load(*project);
    renderer->setResolution(project->getResolution());
    renderer->setCutWorkpiece(simulationManager->getCutWorkpiece());
    view->setToolPath(simulationManager->getToolPath());
  } CBANG_CATCH_ERROR;
}


void GUIApp::redraw() {
  dirty = true;
}


void GUIApp::newProject(bool edit) {
  project = new Project(options, simulationManager->getToolTable());

  // Reset
  simulationManager->reset();
  renderer->setCutWorkpiece(0);
  renderer->stopRender();
  view->setToolPath(0);

  if (edit) editProject();
}


void GUIApp::openProject(const string &_filename) {
  string filename = _filename;

  if (filename.empty()) {
    getFileDialog().setTitle("Open Project");
    if (!getFileDialog().open(window)) return;
    filename = getFileDialog().getFilename();
  }

  LOG_INFO(1, "Opening project " << filename);

  // Check if the file appears to be XML
  SmartPointer<iostream> stream = SystemUtilities::open(filename, ios::in);
  bool xml = false;
  while (!xml) {
    int c = stream->peek();
    if (c == '<') xml = true;
    else if (isspace(c)) stream->get(); // Next
    else break; // Not XML
  }

  if (xml) {
    project = new Project(options, simulationManager->getToolTable(), filename);
    setProjectFile(filename);

  } else {
    // Assume G-Code and create a new project with the file
    newProject(false);
    project->addFile(filename);
    setProjectFile(SystemUtilities::splitExt(filename)[0] + ".xml");
  }

  reload();
  view->getToolPathView()->setByRatio(1);
}


void GUIApp::saveProject(bool saveAs) {
  string filename = project->getFilename();

  if (saveAs || filename.empty()) {
    getFileDialog().setTitle("Save Project");
    if (!getFileDialog().save(window, "OpenSCAM Projects", "*.xml")) return;
    filename = getFileDialog().getFilename();
  }

  project->save(filename);
  setProjectFile(filename);
}


void GUIApp::editProject() {
  projectDialog->edit(project);
}


void GUIApp::exportSTL() {
  SmartPointer<Surface> surface = view->renderer->getSurface();
  if (surface.isNull()) {
    warning("No surface to export");
    return;
  }

  getFileDialog().setTitle("Export STL");
  if (!getFileDialog().save(window, "STL Files", "*.stl")) return;
  string filename = getFileDialog().getFilename();

  STL stl("Cut Workpiece");
  surface->exportSTL(stl);

  SmartPointer<iostream> stream = SystemUtilities::open(filename, ios::out);
  stl.setBinary(true);
  stl.write(*stream);
}


void GUIApp::quit() {
  LOG_INFO(1, "Exiting");
  requestExit();
  gtk_main_quit();
}


bool GUIApp::question(const string &message) {
  GtkWidget *dialog =
    gtk_message_dialog_new(window, GTK_DIALOG_DESTROY_WITH_PARENT,
                           GTK_MESSAGE_QUESTION, GTK_BUTTONS_OK_CANCEL,
                           "%s", message.c_str());

  bool result = gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_OK;
  gtk_widget_destroy(dialog);

  return result;
}


void GUIApp::warning(const string &message) {
  LOG_WARNING(message);

  GtkWidget *dialog =
    gtk_message_dialog_new(window, GTK_DIALOG_DESTROY_WITH_PARENT,
                           GTK_MESSAGE_WARNING, GTK_BUTTONS_OK,
                           "%s", message.c_str());

  gtk_dialog_run(GTK_DIALOG(dialog));
  gtk_widget_destroy(dialog);
}


void GUIApp::error(const string &message) {
  LOG_ERROR(message);

  GtkWidget *dialog =
    gtk_message_dialog_new(window, GTK_DIALOG_DESTROY_WITH_PARENT,
                           GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, "%s",
                           message.c_str());

  gtk_dialog_run(GTK_DIALOG(dialog));
  gtk_widget_destroy(dialog);
}


void GUIApp::on_render_mode_changed(GtkComboBox *widget) {
  GtkTreeIter iter;

  if (!gtk_combo_box_get_active_iter(widget, &iter)) return;

  GtkTreeModel *model = gtk_combo_box_get_model(widget);
  GValue value = {0,};
  gtk_tree_model_get_value(model, &iter, 1, &value);

  renderer->setMode((RenderMode::enum_t)g_value_get_uint(&value));
  g_value_unset(&value);
}


void GUIApp::on_program_scrollbar_changed(GtkRange *range) {
  double value = gtk_range_get_value(range);
  view->path->setByRatio(value);
  redraw();
}


void GUIApp::on_check_button_toggled(GtkToggleButton *widget) {
  string name = gtk_buildable_get_name((GtkBuildable *)widget);
  bool active = gtk_toggle_button_get_active(widget);

  int flag = 0;
  if (name == "wire_check") flag = View::WIRE_FLAG;
  else if (name == "bounds_check") view->setShowBounds(active);
  else if (name == "axes_check") view->setShowAxes(active);
  else if (name == "path_check") flag = View::SHOW_PATH_FLAG;
  else if (name == "tool_check") flag = View::SHOW_TOOL_FLAG;
  else if (name == "surface_check") flag = View::SHOW_SURFACE_FLAG;
  else if (name == "bbtree_check") flag = View::SHOW_BBTREE_FLAG;
  else if (name == "normals_check") flag = View::SHOW_NORMALS_FLAG;
  else if (name == "smooth_check") renderer->toggleSmooth();

  view->setFlag(flag, active);

  redraw();
}


void GUIApp::on_tool_button_clicked(GtkWidget *widget) {
  string name = gtk_buildable_get_name((GtkBuildable *)widget);

  if (name == "new_button") newProject();
  else if (name == "open_button") openProject();
  else if (name == "save_button") saveProject();
  else if (name == "project_button") editProject();
  else if (name == "export_stl_button") exportSTL();
  else if (name == "about_button") aboutDialog->show();
  else if (name == "play_button") {
    view->toggleFlag(View::PLAY_FLAG);

    const char *stock = view->isFlagSet(View::PLAY_FLAG) ?
      GTK_STOCK_MEDIA_PAUSE : GTK_STOCK_MEDIA_PLAY;
    gtk_tool_button_set_stock_id(playButton, stock);

  } else if (name == "slower_button") view->decSpeed();
  else if (name == "faster_button") view->incSpeed();
  else if (name == "start_button") view->path->setByRatio(0);
  else if (name == "end_button") view->path->setByRatio(1);
  else if (name == "direction_button") view->changeDirection();
  else if (name == "render_reload_button") rerender();
  else if (name == "render_stop_button") renderer->stopRender();
  else if (name == "isometric_button") view->resetView('p');
  else if (name == "front_button") view->resetView('F');
  else if (name == "back_button") view->resetView('B');
  else if (name == "left_button") view->resetView('l');
  else if (name == "right_button") view->resetView('r');
  else if (name == "top_button") view->resetView('t');
  else if (name == "bottom_button") view->resetView('b');
  else LOG_INFO(1, name << " not yet implemented");

  redraw();
}


void GUIApp::on_menu_item_activate(GtkMenuItem *item) {
  string name = gtk_buildable_get_name((GtkBuildable *)item);
  LOG_DEBUG(3, name);

  if (name == "new_item") newProject();
  else if (name == "open_item") openProject();
  else if (name == "save_item") saveProject();
  else if (name == "save_as_item") saveProject(true);
  else if (name == "export_stl_item") exportSTL();
  else if (name == "quit_item") quit();
  else if (name == "project_item") editProject();
  else if (name == "about_item") aboutDialog->show();
  else LOG_INFO(1, name << " not yet implemented");

  redraw();
}


bool GUIApp::on_glarea_mouse_press(GtkWidget *widget, GdkEventButton *event) {
  LOG_DEBUG(3, "on_glarea_mouse_press(button=" << event->button << ", "
            << (event->type == GDK_BUTTON_PRESS ? "press" : "release")
            << ", x=" << event->x << ", y=" << event->y << ")");

  switch (event->type) {
  case GDK_BUTTON_PRESS:
    switch (event->button) {
    case 1: view->startRotation(event->x, event->y); break;
    case 2: view->startTranslation(event->x, event->y); break;
    }
  default: break;
  }

  return true;
}


bool GUIApp::on_glarea_mouse_motion(GtkWidget *widget, GdkEventMotion *event) {
  LOG_DEBUG(5, "on_glarea_mouse_motion(x=" << event->x
            << ", y=" << event->y << ")");

  if (event->state & GDK_BUTTON1_MASK) view->updateRotation(event->x, event->y);
  if (event->state & GDK_BUTTON2_MASK)
    view->updateTranslation(event->x, event->y);

  if (event->state & (GDK_BUTTON1_MASK | GDK_BUTTON2_MASK)) redraw();

  return true;
}


bool GUIApp::on_glarea_mouse_scroll(GtkWidget *widget, GdkEventScroll *event) {
  bool changed = true;
  switch (event->direction) {
  case GDK_SCROLL_UP: view->zoomIn(); break;
  case GDK_SCROLL_DOWN: view->zoomOut(); break;
  default: changed = false; break;
  }

  if (changed) redraw();
  return changed;
}


void GUIApp::on_glarea_realize(GtkWidget *widget) {
  LOG_DEBUG(3, "glarea.realize()");

  if (!widget->window) return;

  GdkGLContext *glcontext = gtk_widget_get_gl_context(widget);
  GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable(widget);

  if (!gdk_gl_drawable_gl_begin(gldrawable, glcontext)) return;

  GLenum err = glewInit();
  if (err != GLEW_OK) THROWS("Initializing GLEW: " << glewGetErrorString(err));

  view->glInit();
  viewer->init();

  gdk_gl_drawable_gl_end(gldrawable);
}


bool GUIApp::on_glarea_configure(GtkWidget *widget, GdkEventConfigure *event) {
  if (!widget->window) return false;

  GdkGLContext *glcontext = gtk_widget_get_gl_context(widget);
  GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable(widget);

  float width = widget->allocation.width;
  float height = widget->allocation.height;

  LOG_DEBUG(3, "glarea.configure(" << width << ", " << height << ')');

  if (!gdk_gl_drawable_gl_begin(gldrawable, glcontext)) return false;

  view->resize(width, height);

  gdk_gl_drawable_gl_end(gldrawable);

  return true;
}


bool GUIApp::on_glarea_expose(GtkWidget *widget, GdkEventExpose *event) {
  if (!widget->window) return false;

  GdkGLContext *glcontext = gtk_widget_get_gl_context(widget);
  GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable(widget);

  if (!gdk_gl_drawable_gl_begin(gldrawable, glcontext)) return false;

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  viewer->draw(*view);

  gdk_gl_drawable_swap_buffers(gldrawable);
  gdk_gl_drawable_gl_end(gldrawable);

  renderTimer.throttle(1.0 / 60); // No faster than X Hz

  return true;
}


bool GUIApp::on_window_delete_event(GtkWindow *window, GdkEvent *event) {
  return false;
}


void GUIApp::on_window_destroy(GtkWindow *window) {
  quit();
}


bool GUIApp::on_idle() {
  try {
    if (project->checkFiles()) {
      reload();
      dirty = true;
    }

    dirty = connectionManager->update() || dirty;
    dirty = view->update() || dirty;
    dirty = renderer->wasUpdated() || dirty;

    if (dirty) {
      gdk_window_invalidate_rect(glArea->window, &glArea->allocation, false);
      dirty = false;
    }

    // Update render progress
    double progress = renderer->getProgress();
    double eta = renderer->getETA();
    gtk_progress_bar_set_fraction(renderProgress, progress);
    if (progress) {
      string s = String::printf("%.2f%% ", progress * 100);
      if (eta) s += TimeInterval(eta).toString();
      gtk_progress_bar_set_text(renderProgress, s.c_str());

    } else gtk_progress_bar_set_text(renderProgress, "");

    // Throttle
    idleTimer.throttle(1.0 / 60); // No faster than 60Hz
  } CBANG_CATCH_ERROR;

  return true; // Keep going
}


bool GUIApp::on_map_event(GtkWidget *widget, GdkEvent *event) {
  if (!idleID)
    idleID =
      g_idle_add_full(GDK_PRIORITY_REDRAW, (GSourceFunc)::on_idle, this, 0);

  return true;
}


bool GUIApp::on_unmap_event(GtkWidget *widget, GdkEvent *event) {
  if (idleID) {
    g_source_remove(idleID);
    idleID = 0;
  }

  return true;
}


void GUIApp::directionUpdate(const std::string &name, bool reverse) {
  gtk_tool_button_set_stock_id(directionButton,
                               reverse ? "gtk-go-back" : "gtk-go-forward");
}
