/******************************************************************************\

    OpenSCAM is an Open-Source CAM software.
    Copyright (C) 2011-2012 Joseph Coffland <joseph@cauldrondevelopment.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

\******************************************************************************/

#ifndef OPENSCAM_GUI_APP_H
#define OPENSCAM_GUI_APP_H

#include "GTKHelper.h"
#include "Builder.h"
#include "ToolDialog.h"
#include "ProjectDialog.h"
#include "FileDialog.h"

#include <openscam/view/View.h>
#include <openscam/value/ValueSet.h>

#include <cbang/SmartPointer.h>
#include <cbang/Application.h>


namespace OpenSCAM {
  class SimulationManager;
  class ConnectionManager;
  class Renderer;
  class Viewer;
  class Project;

  class GUIApp : public cb::Application {
    cb::SmartPointer<Project> project;

    ValueSet valueSet;
    cb::SmartPointer<SimulationManager> simulationManager;
    cb::SmartPointer<ConnectionManager> connectionManager;
    cb::SmartPointer<Renderer> renderer;
    cb::SmartPointer<View> view;
    cb::SmartPointer<Viewer> viewer;
    cb::SmartPointer<Builder> builder;

    bool dirty;

    int idleID;
    GtkWindow *window;
    GtkWindow *glWindow;
    GtkWidget *glArea;
    GtkToolButton *playButton;
    GtkToolButton *directionButton;
    GtkProgressBar *renderProgress;
    cb::SmartPointer<Dialog> aboutDialog;
    cb::SmartPointer<FileDialog> fileDialog;
    cb::SmartPointer<ProjectDialog> projectDialog;

    cb::Timer renderTimer;
    cb::Timer idleTimer;

  public:
    GUIApp();
    ~GUIApp();

    void setProjectFile(const std::string &filename);

    const cb::SmartPointer<SimulationManager> &getSimulationManager() const
    {return simulationManager;}
    const cb::SmartPointer<ConnectionManager> &getConnectionManager() const
    {return connectionManager;}
    const cb::SmartPointer<Renderer> &getRenderer() const {return renderer;}
    const cb::SmartPointer<View> &getView() const {return view;}
    const cb::SmartPointer<Builder> &getBuilder() {return builder;}

    FileDialog &getFileDialog() {return *fileDialog;}
    ProjectDialog &getProjectDialog() {return *projectDialog;}

    void loadImage(const char *name, const guint8 data[]);

    // From Application
    int init(int argc, char *argv[]);
    void run();

    void rerender();
    void reload();
    void redraw();
    void newProject(bool edit = true);
    void openProject(const std::string &filename = std::string());
    void saveProject(bool saveas = false);
    void editProject();
    void exportSTL();
    void quit();

    bool question(const std::string &message);
    void warning(const std::string &message);
    void error(const std::string &message);

    // Events
    void on_render_mode_changed(GtkComboBox *widget);

    void on_program_scrollbar_changed(GtkRange *range);

    void on_check_button_toggled(GtkToggleButton *widget);
    void on_tool_button_clicked(GtkWidget *widget);

    void on_menu_item_activate(GtkMenuItem *item);

    bool on_glarea_mouse_press(GtkWidget *widget, GdkEventButton *event);
    bool on_glarea_mouse_motion(GtkWidget *widget, GdkEventMotion *event);
    bool on_glarea_mouse_scroll(GtkWidget *widget, GdkEventScroll *event);

    void on_glarea_realize(GtkWidget *widget);
    bool on_glarea_configure(GtkWidget *widget, GdkEventConfigure *event);
    bool on_glarea_expose(GtkWidget *widget, GdkEventExpose *event);

    bool on_window_delete_event(GtkWindow *window, GdkEvent *event);
    void on_window_destroy(GtkWindow *window);

    bool on_idle();
    bool on_map_event(GtkWidget *widget, GdkEvent *event);
    bool on_unmap_event(GtkWidget *widget, GdkEvent *event);

    // Observers
    void directionUpdate(const std::string &name, bool reverse);
  };
}

#endif // OPENSCAM_GUI_APP_H

