/******************************************************************************\

    OpenSCAM is an Open-Source CAM software.
    Copyright (C) 2011-2012 Joseph Coffland <joseph@cauldrondevelopment.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

\******************************************************************************/

#include "ProjectDialog.h"

#include "GUIApp.h"
#include "Builder.h"
#include "GTKHelper.h"

#include <openscam/cutsim/Project.h>
#include <openscam/sim/Controller.h>
#include <openscam/sim/ToolTable.h>

#include <cbang/String.h>
#include <cbang/os/SystemUtilities.h>

using namespace std;
using namespace cb;
using namespace OpenSCAM;


GTK_SIGNAL1V(on_resolution_changed, GtkWidget *, ProjectDialog);
GTK_SIGNAL3V(on_tool_table_row_activated, GtkTreeView *, GtkTreePath *,
             GtkTreeViewColumn *, ProjectDialog);

GTK_SIGNAL1V(on_nc_files_add, GtkWidget *, ProjectDialog);
GTK_SIGNAL1V(on_nc_files_remove, GtkWidget *, ProjectDialog);
GTK_SIGNAL3V(on_nc_file_edited, GtkWidget *, char *, char *, ProjectDialog);

GTK_SIGNAL1V(on_tool_table_add, GtkWidget *, ProjectDialog);
GTK_SIGNAL1V(on_tool_table_remove, GtkWidget *, ProjectDialog);


ProjectDialog::ProjectDialog(GUIApp &app) :
  Dialog(*app.getBuilder(), "project"), app(app), toolDialog(app) {

  gtk_dialog_add_buttons(dialog, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                         GTK_STOCK_OK, GTK_RESPONSE_OK, (void *)0);

  Builder &builder = *app.getBuilder();

  resolutionCombo = builder.get<GtkComboBox>("render_resolution_combo");
  resolutionSpin = builder.get<GtkSpinButton>("render_resolution_spin");

  ncFileTree = builder.get<GtkTreeView>("nc_file_tree");
  ncFileStore = builder.get<GtkListStore>("nc_file_store");

  toolTableTree = builder.get<GtkTreeView>("tool_table_tree");
  toolTableStore = builder.get<GtkListStore>("tool_table_list");

  workpieceTable = builder.get<GtkWidget>("workpiece_table");
  autoWorkpieceRadio = builder.get<GtkRadioButton>("automatic_radio");
  cuboidWorkpieceRadio = builder.get<GtkRadioButton>("cuboid_radio");

  wpMarginSpin = builder.get<GtkSpinButton>("workpiece_margin_spin");
  minXSpin = builder.get<GtkSpinButton>("workpiece_min_x_spin");
  minYSpin = builder.get<GtkSpinButton>("workpiece_min_y_spin");
  minZSpin = builder.get<GtkSpinButton>("workpiece_min_z_spin");
  maxXSpin = builder.get<GtkSpinButton>("workpiece_max_x_spin");
  maxYSpin = builder.get<GtkSpinButton>("workpiece_max_y_spin");
  maxZSpin = builder.get<GtkSpinButton>("workpiece_max_z_spin");

  builder.connect("render_resolution_spin", "changed",
                  G_CALLBACK(::on_resolution_changed), this);
  builder.connect("nc_files_add_button", "clicked",
                  G_CALLBACK(::on_nc_files_add), this);
  builder.connect("nc_files_remove_button", "clicked",
                  G_CALLBACK(::on_nc_files_remove), this);
  builder.connect("nc_file_cellrenderer", "edited",
                  G_CALLBACK(::on_nc_file_edited), this);

  builder.connect("tool_table_add_button", "clicked",
                  G_CALLBACK(::on_tool_table_add), this);
  builder.connect("tool_table_remove_button", "clicked",
                  G_CALLBACK(::on_tool_table_remove), this);
  builder.connect("tool_table_tree", "row-activated",
                  G_CALLBACK(::on_tool_table_row_activated), this);
}


ProjectDialog::~ProjectDialog() {} // Hide member destructors


bool ProjectDialog::getAutomaticWorkpiece() {
  return gtk_toggle_button_get_active((GtkToggleButton *)autoWorkpieceRadio);
}


void ProjectDialog::setAutomaticWorkpiece(bool x) {
  if (getAutomaticWorkpiece() != x)
    gtk_toggle_button_set_active((GtkToggleButton *)autoWorkpieceRadio, x);
}


void ProjectDialog::load(const SmartPointer<Project> &project) {
  this->project = project;
  loading = true;

  // Render Resolution
  gtk_spin_button_set_value(resolutionSpin, project->getResolution());
  gtk_combo_box_set_active(resolutionCombo, project->getResolutionMode());

  // NC Files
  vector<string> files = project->getRelativeFiles();

  gtk_list_store_clear(ncFileStore);
  for (unsigned i = 0; i < files.size(); i++) {
    GtkTreeIter iter;
    gtk_list_store_append(ncFileStore, &iter);
    gtk_list_store_set(ncFileStore, &iter, 0, files[i].c_str(), -1);
  }

  // Tools
  gtk_list_store_clear(toolTableStore);
  ToolTable &tools = project->getToolTable();
  for (ToolTable::iterator it = tools.begin(); it != tools.end(); it++)
    if (it->second->getNumber()) addTool(it->second);

  // Workpiece
  if (project->getAutomaticWorkpiece())
    gtk_toggle_button_set_active((GtkToggleButton *)autoWorkpieceRadio, true);
  else
    gtk_toggle_button_set_active((GtkToggleButton *)cuboidWorkpieceRadio, true);

  double margin = project->getWorkpieceMargin();
  gtk_spin_button_set_value(wpMarginSpin, margin);

  Rectangle3D wpBounds = project->getWorkpieceBounds();
  Vector3D wpMin = wpBounds.getMin();
  Vector3D wpMax = wpBounds.getMax();

  gtk_spin_button_set_value(minXSpin, wpMin.x());
  gtk_spin_button_set_value(minYSpin, wpMin.y());
  gtk_spin_button_set_value(minZSpin, wpMin.z());
  gtk_spin_button_set_value(maxXSpin, wpMax.x());
  gtk_spin_button_set_value(maxYSpin, wpMax.y());
  gtk_spin_button_set_value(maxZSpin, wpMax.z());

  loading = false;
}


void ProjectDialog::edit(const SmartPointer<Project> &project) {
  load(project);
  show();
}


void ProjectDialog::addTool(const SmartPointer<Tool> &tool) {
  GtkTreeIter iter;
  gtk_list_store_append(toolTableStore, &iter);

  double scale = tool->getUnits() == ToolUnits::UNITS_INCH ? 1.0 / 25.4 : 1;

  string units = tool->getUnits().toString();
  string shape = tool->getShape().toString();
  gtk_list_store_set(toolTableStore, &iter,
                     0, tool->getNumber(),
                     1, units.c_str(),
                     2, shape.c_str(),
                     3, tool->getLength() * scale,
                     4, 2 * tool->getRadius() * scale,
                     5, tool->getSnubDiameter() * scale,
                     6, tool->getDescription().c_str(),
                     7, tool->getText().c_str(),
                     -1);
}


void ProjectDialog::saveTool() {
  if (newTool) gtk_list_store_append(toolTableStore, &currentTool);

  string units = toolDialog.getUnits().toString();
  string shape = toolDialog.getShape().toString();
  gtk_list_store_set(toolTableStore, &currentTool,
                     0, toolDialog.getNumber(),
                     1, units.c_str(),
                     2, shape.c_str(),
                     3, toolDialog.getLength(),
                     4, toolDialog.getDiameter(),
                     5, toolDialog.getSnubDiameter(),
                     6, toolDialog.getDescription().c_str(),
                     -1);

  Tool tool;
  storeTool(currentTool, tool);
  gtk_list_store_set(toolTableStore, &currentTool, 7, tool.getText().c_str(),
                     -1);
}


void ProjectDialog::storeTool(GtkTreeIter iter, Tool &tool) {
  gchar *units = 0;
  gchar *shape = 0;
  gdouble length;
  gdouble diameter;
  gdouble snubDiameter;
  gchar *description = 0;
  gtk_tree_model_get((GtkTreeModel *)toolTableStore, &iter, 1, &units,
                     2, &shape, 3, &length, 4, &diameter, 5, &snubDiameter,
                     6, &description, -1);

  tool.setUnits(ToolUnits::parse(units));
  double scale = tool.getUnits() == ToolUnits::UNITS_INCH ? 25.4 : 1;
  tool.setShape(ToolShape::parse(shape));
  tool.setLength(length * scale);
  tool.setRadius(diameter / 2.0 * scale);
  tool.setSnubDiameter(snubDiameter * scale);
  tool.setDescription(description);

  g_free(units);
  g_free(shape);
  g_free(description);
}


void ProjectDialog::removeListSelection(GtkTreeView *tree) {
  GtkTreeSelection *selection = gtk_tree_view_get_selection(tree);
  GtkTreeModel *model = 0;
  GtkTreeIter iter;
  if (!gtk_tree_selection_get_selected(selection, &model, &iter)) return;

  gtk_list_store_remove((GtkListStore *)model, &iter);
}


void ProjectDialog::on_resolution_changed(GtkWidget *widget) {
  LOG_INFO(1, "on_resolution_changed() " << loading);
  if (loading) return;
  gtk_combo_box_set_active(resolutionCombo, ResolutionMode::RESOLUTION_MANUAL);
}


void ProjectDialog::on_tool_table_row_activated(GtkTreeView *tree,
                                                GtkTreePath *path,
                                                GtkTreeViewColumn *column) {
  GtkTreeModel *model = gtk_tree_view_get_model(tree);
  if (!gtk_tree_model_get_iter(model, &currentTool, path)) return;

  guint number;
  gchar *units;
  gchar *shape;
  gdouble length;
  gdouble diameter;
  gdouble snubDiameter;
  gchar *description;
  gtk_tree_model_get(model, &currentTool, 0, &number, 1, &units, 2, &shape,
                     3, &length, 4, &diameter, 5, &snubDiameter,
                     6, &description, -1);

  toolDialog.setNumber(number);
  toolDialog.setUnits(ToolUnits::parse(units));
  toolDialog.setShape(ToolShape::parse(shape));
  toolDialog.setLength(length);
  toolDialog.setDiameter(diameter);
  toolDialog.setSnubDiameter(snubDiameter);
  toolDialog.setDescription(description);

  g_free(units);
  g_free(shape);
  g_free(description);

  newTool = false;

  toolDialog.show();
}


void ProjectDialog::on_nc_files_add(GtkWidget *widget) {
  if (app.getFileDialog().open(getWindow(), "GCode files",
                               "*.nc *.ngc *.gcode")) {
    GtkTreeIter iter;
    gtk_list_store_append(ncFileStore, &iter);

    string filename = app.getFileDialog().getFilename();

    // Make relative path
    string relativePath = project->getFilename();
    if (relativePath.empty()) relativePath = SystemUtilities::getcwd();
    else relativePath = SystemUtilities::dirname(relativePath);
    filename = SystemUtilities::relative(relativePath, filename, 4);

    LOG_DEBUG(1, "Adding file " << filename);

    gtk_list_store_set(ncFileStore, &iter, 0, filename.c_str(), -1);
  }
}


void ProjectDialog::on_nc_files_remove(GtkWidget *widget) {
  removeListSelection(ncFileTree);
}


void ProjectDialog::on_tool_table_add(GtkWidget *widget) {
  newTool = true;
  toolDialog.show();
}


void ProjectDialog::on_nc_file_edited(GtkWidget *widget, char *_path,
                                      char *text) {
  GtkTreePath *path = gtk_tree_path_new_from_string(_path);
  GtkTreeIter iter;
  gtk_tree_model_get_iter((GtkTreeModel *)ncFileStore, &iter, path);
  gtk_tree_path_free(path);

  gtk_list_store_set(ncFileStore, &iter, 0, text, -1);
}


void ProjectDialog::on_tool_table_remove(GtkWidget *widget) {
  removeListSelection(toolTableTree);
}


void ProjectDialog::accept() {
  // Check that more than one tool with the same number does not exist
  set<unsigned> numbers;
  GtkTreeIter iter;
  GtkTreeModel *model = (GtkTreeModel *)toolTableStore;
  if (gtk_tree_model_get_iter_first(model, &iter)) {
    while (true) {
      GValue value = {0,};
      gtk_tree_model_get_value(model, &iter, 0, &value);

      unsigned number = g_value_get_uint(&value);
      g_value_unset(&value);

      if (!number) {
        app.error("Cannot define tool zero");
        return;
      }

      if (!numbers.insert(number).second) {
        app.error(SSTR("More than one tool with number " << number
                       << " exists"));
        return;
      }

      if (!gtk_tree_model_iter_next(model, &iter)) break;
    }
  }

  // Save Render Resolution
  project->setResolutionMode
    ((ResolutionMode::enum_t)gtk_combo_box_get_active(resolutionCombo));
  project->setResolution(gtk_spin_button_get_value(resolutionSpin));

  // Save NC Files
  vector<string> files;
  model = (GtkTreeModel *)ncFileStore;
  if (gtk_tree_model_get_iter_first(model, &iter)) {
    while (true) {
      gchar *filename;
      gtk_tree_model_get(model, &iter, 0, &filename, -1);

      files.push_back(filename);

      g_free(filename);
      if (!gtk_tree_model_iter_next(model, &iter)) break;
    }
  }
  project->setFiles(files);

  // Save Tool Table
  ToolTable &tools = project->getToolTable();
  tools.clear();
  model = (GtkTreeModel *)toolTableStore;
  if (gtk_tree_model_get_iter_first(model, &iter)) {
    while (true) {
      guint number;
      gtk_tree_model_get(model, &iter, 0, &number, -1);
      SmartPointer<Tool> tool = new Tool(number);
      storeTool(iter, *tool);
      tools.add(tool);

      if (!gtk_tree_model_iter_next(model, &iter)) break;
    }
  }

  // Save Workpiece
  project->setAutomaticWorkpiece(getAutomaticWorkpiece());
  double margin = gtk_spin_button_get_value(wpMarginSpin);
  project->setWorkpieceMargin(margin);
  Rectangle3R wpBounds(Vector3D(gtk_spin_button_get_value(minXSpin),
                                gtk_spin_button_get_value(minYSpin),
                                gtk_spin_button_get_value(minZSpin)),
                       Vector3D(gtk_spin_button_get_value(maxXSpin),
                                gtk_spin_button_get_value(maxYSpin),
                                gtk_spin_button_get_value(maxZSpin)));
  project->setWorkpieceBounds(wpBounds);

  // Finish up
  app.reload();
  hide();
}
