/******************************************************************************\

    OpenSCAM is an Open-Source CAM software.
    Copyright (C) 2011-2012 Joseph Coffland <joseph@cauldrondevelopment.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

\******************************************************************************/

#include "ToolDialog.h"

#include "GUIApp.h"
#include "Builder.h"

#include <openscam/sim/Tool.h>

#include <cbang/String.h>
#include <cbang/util/SmartToggle.h>

using namespace std;
using namespace cb;
using namespace OpenSCAM;


GTK_SIGNAL1V(on_tool_units_changed, GtkWidget *, ToolDialog);
GTK_SIGNAL1V(on_tool_changed, GtkWidget *, ToolDialog);


ToolDialog::ToolDialog(GUIApp &app) :
  Dialog(*app.getBuilder(), "tool"), app(app), busy(false), width(250),
  height(270) {

  gtk_dialog_add_buttons(dialog, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                         GTK_STOCK_OK, GTK_RESPONSE_OK, (void *)0);

  Builder &builder = *app.getBuilder();

  numberSpin = builder.get<GtkSpinButton>("tool_number");
  unitsCombo = builder.get<GtkComboBox>("tool_units");
  shapeCombo = builder.get<GtkComboBox>("tool_shape");
  lengthSpin = builder.get<GtkSpinButton>("tool_length");
  diameterSpin = builder.get<GtkSpinButton>("tool_diameter");
  snubDiameterSpin = builder.get<GtkSpinButton>("snub_diameter");
  snubDiameterLabel = builder.get<GtkLabel>("snub_diameter_label");
  descriptionEntry = builder.get<GtkEntry>("tool_description");
  toolImage = builder.get<GtkImage>("tool_image");

  builder.connect("tool_units", "changed",
                  G_CALLBACK(::on_tool_units_changed), this);
  builder.connect("tool_shape", "changed", G_CALLBACK(::on_tool_changed), this);
  builder.connect("tool_length", "changed",
                  G_CALLBACK(::on_tool_changed), this);
  builder.connect("tool_diameter", "changed",
                  G_CALLBACK(::on_tool_changed), this);
  builder.connect("snub_diameter", "changed",
                  G_CALLBACK(::on_tool_changed), this);

  pixmap = gdk_pixmap_new(0, width, height, 24);
}


unsigned ToolDialog::getNumber() {
  return gtk_spin_button_get_value_as_int(numberSpin);
}


void ToolDialog::setNumber(unsigned number) {
  gtk_spin_button_set_value(numberSpin, number);
}


ToolUnits ToolDialog::getUnits() {
  GtkTreeIter iter;

  if (!gtk_combo_box_get_active_iter(unitsCombo, &iter))
    return ToolUnits::UNITS_MM;

  GtkTreeModel *model = gtk_combo_box_get_model(unitsCombo);
  GValue value = {0,};
  gtk_tree_model_get_value(model, &iter, 0, &value);

  ToolUnits units = ToolUnits::parse(g_value_get_string(&value));
  g_value_unset(&value);

  return units;
}


void ToolDialog::setUnits(ToolUnits units) {
  lastUnits = units;

  string unitsStr = units.toString();
  GtkTreeModel *model = gtk_combo_box_get_model(unitsCombo);
  GtkTreeIter iter;

  if (!gtk_tree_model_get_iter_first(model, &iter)) return;

  while (true) {
    GValue value = {0,};
    gtk_tree_model_get_value(model, &iter, 0, &value);

    if (String::toUpper(g_value_get_string(&value)) == unitsStr) {
      gtk_combo_box_set_active_iter(unitsCombo, &iter);
      break;
    }

    g_value_unset(&value);
    if (!gtk_tree_model_iter_next(model, &iter)) break;
  }
}


ToolShape ToolDialog::getShape() {
  GtkTreeIter iter;

  if (!gtk_combo_box_get_active_iter(shapeCombo, &iter))
    return ToolShape::TS_CYLINDRICAL;

  GtkTreeModel *model = gtk_combo_box_get_model(shapeCombo);
  GValue value = {0,};
  gtk_tree_model_get_value(model, &iter, 0, &value);

  ToolShape shape = ToolShape::parse(g_value_get_string(&value));
  g_value_unset(&value);

  return shape;
}


void ToolDialog::setShape(ToolShape shape) {
  string shapeStr = shape.toString();
  GtkTreeModel *model = gtk_combo_box_get_model(shapeCombo);
  GtkTreeIter iter;

  if (!gtk_tree_model_get_iter_first(model, &iter)) return;

  while (true) {
    GValue value = {0,};
    gtk_tree_model_get_value(model, &iter, 0, &value);

    if (String::toUpper(g_value_get_string(&value)) == shapeStr) {
      gtk_combo_box_set_active_iter(shapeCombo, &iter);
      break;
    }

    g_value_unset(&value);
    if (!gtk_tree_model_iter_next(model, &iter)) break;
  }
}


double ToolDialog::getLength() {
  return gtk_spin_button_get_value(lengthSpin);
}


void ToolDialog::setLength(double length) {
  gtk_spin_button_set_value(lengthSpin, length);
}


double ToolDialog::getDiameter() {
  return gtk_spin_button_get_value(diameterSpin);
}


void ToolDialog::setDiameter(double diameter) {
  gtk_spin_button_set_value(diameterSpin, diameter);
}


double ToolDialog::getSnubDiameter() {
  return gtk_spin_button_get_value(snubDiameterSpin);
}


void ToolDialog::setSnubDiameter(double diameter) {
  gtk_spin_button_set_value(snubDiameterSpin, diameter);
}

string ToolDialog::getDescription() {
  return gtk_entry_get_text(descriptionEntry);
}


void ToolDialog::setDescription(const string &description) {
  gtk_entry_set_text(descriptionEntry, description.c_str());
}


void ToolDialog::on_tool_units_changed(GtkWidget *widget) {
  if (busy) return;
  SmartToggle toggle(busy);

  ToolUnits newUnits = getUnits();
  if (newUnits == lastUnits) return;
  lastUnits = newUnits;

  double scale = newUnits == ToolUnits::UNITS_MM ? 25.4 : 1.0 / 25.4;
  setLength(getLength() * scale);
  setDiameter(getDiameter() * scale);
  setSnubDiameter(getSnubDiameter() * scale);
}


static void draw_guide(cairo_t *cr, double width, double x, double y,
                       bool up, const char *text) {
  cairo_set_line_width(cr, 1);
  cairo_move_to(cr, x - width / 2.0, y + (up ? 0 : 15));
  cairo_rel_line_to(cr, 0, up ? 3 : -3);
  cairo_rel_line_to(cr, width, 0);
  cairo_rel_line_to(cr, 0, up ? -3 : 3);
  cairo_stroke(cr);

  cairo_set_font_size(cr, 10);
  cairo_move_to(cr, 5, y + (up ? 13 : 8));
  cairo_show_text(cr, text);

  cairo_rel_move_to(cr, 2, -3);
  cairo_line_to(cr, x, y + (up ? 10 : 5));
  cairo_rel_line_to(cr, 0, up ? -7 : 7);
  cairo_stroke(cr);
}


void ToolDialog::on_tool_changed(GtkWidget *widget) {
  if (busy) return;
  SmartToggle toggle(busy);

  ToolShape shape = getShape();
  double length = getLength();
  double diameter = getDiameter();
  double snubDiameter = shape == ToolShape::TS_SNUBNOSE ? getSnubDiameter() : 0;

  gtk_widget_set_visible((GtkWidget *)snubDiameterSpin,
                         shape == ToolShape::TS_SNUBNOSE);
  gtk_widget_set_visible((GtkWidget *)snubDiameterLabel,
                         shape == ToolShape::TS_SNUBNOSE);

  const double bg = 1.0;
  const double fg = 0.2;

  cairo_t *cr = gdk_cairo_create(pixmap);

  // Blank background
  cairo_set_source_rgba(cr, bg, bg, bg, 1);
  cairo_rectangle(cr, 0, 0, width, height);
  cairo_fill(cr);
  cairo_set_source_rgba(cr, fg, fg, fg, 1);

  // Margin
  cairo_save(cr);
  cairo_translate(cr, 25, 25);
  double w = 200;
  double h = 200;

  // Scale
  double maxDim = diameter;
  if (maxDim < length) maxDim = length;
  double scale = h / maxDim;
  cairo_scale(cr, scale, scale);
  w /= scale;
  h /= scale;

  // Cutter
  double x = (w - diameter) / 2.0;
  double y = (h - length) / 2.0;
  switch (shape) {
  case ToolShape::TS_CYLINDRICAL:
    cairo_rectangle(cr, x, y, diameter, length);
    break;

  case ToolShape::TS_SNUBNOSE:
  case ToolShape::TS_CONICAL:
    cairo_move_to(cr, x, y);
    if (diameter) cairo_rel_line_to(cr, diameter, 0);
    cairo_rel_line_to(cr, (snubDiameter - diameter) / 2, length);
    if (snubDiameter) cairo_rel_line_to(cr, -snubDiameter, 0);
    cairo_close_path(cr);
    break;

  case ToolShape::TS_BALLNOSE: {
    double radius = diameter / 2.0;
    cairo_move_to(cr, x, y);
    cairo_rel_line_to(cr, diameter, 0);
    cairo_arc(cr, w / 2.0, y + length - radius, radius, 0, M_PI);
    cairo_close_path(cr);
    break;
  }

  case ToolShape::TS_SPHEROID:
    cairo_save(cr);
    cairo_translate(cr, w / 2.0, y + length / 2.0);
    cairo_scale(cr, diameter / 2.0, length / 2.0);
    cairo_arc(cr, 0, 0, 1, 0, 2 * M_PI);
    cairo_restore(cr);
    break;
  }

  cairo_fill(cr);

  cairo_restore(cr);
  w *= scale;
  h *= scale;

  // Guides
  x = width / 2.0;
  y = h + 30;
  draw_guide(cr, diameter * scale, x, y, true, "Diameter");
  if (shape == ToolShape::TS_SNUBNOSE)
    draw_guide(cr, snubDiameter * scale, x, y + 20, true, "Snub Diameter");

  cairo_rotate(cr, M_PI / 2.0);
  cairo_translate(cr, 25, -5);
  draw_guide(cr, length * scale, h / 2.0, -15, true, "Length");

  gtk_image_set_from_pixmap(toolImage, pixmap, 0);

  cairo_destroy(cr);
}


void ToolDialog::accept() {
  if (!getLength()) {app.error("Length cannot be zero"); return;}
  if (!getDiameter()) {app.error("Shaft diameter cannot be zero"); return;}

  try {
    app.getProjectDialog().saveTool();
    hide();

  } catch (const Exception &e) {
    app.error(e.getMessage());
  }
}
