/******************************************************************************\

    OpenSCAM is an Open-Source CAM software.
    Copyright (C) 2011-2012 Joseph Coffland <joseph@cauldrondevelopment.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

\******************************************************************************/

#ifndef OPENSCAM_CONTROLLER_H
#define OPENSCAM_CONTROLLER_H

#include "ToolTable.h"
#include "Machine.h"
#include "Addresses.h"
#include "CoordinateSystem.h"

#include <openscam/cutsim/Move.h>
#include <openscam/gcode/VarTypes.h>
#include <openscam/gcode/ModalGroup.h>

#include <cbang/LocationRange.h>

#include <map>
#include <string>
#include <ostream>


namespace cb {class Options;}

namespace OpenSCAM {
  class Code;
  class Entity;

  struct EndProgram {}; // Exception

  // TODO Rename this to Controller
  class Controller : public VarTypes, public ModalGroup {
  public:
    typedef enum {
      XY_PLANE,
      ZX_PLANE,
      YZ_PLANE,
      UV_PLANE,
      WU_PLANE,
      VW_PLANE,
    } plane_t;


    typedef enum {
      DIR_OFF,
      DIR_CLOCKWISE,
      DIR_COUNTERCLOCKWISE,
    } dir_t;


    typedef enum {
      EXACT_PATH_MODE,
      EXACT_STOP_MODE,
      CONTINUOUS_MODE,
    } path_mode_t;


    typedef enum {
      UNITS_PER_MINUTE_MODE,
      UNITS_PER_REVOLUTION_MODE,
      INVERSE_TIME_MODE,
    } feed_mode_t;


    typedef enum {
      SPINDLE_RPM_MODE,
      SPINDLE_CSS_MODE,
    } spindle_mode_t;


    typedef enum {
      RETURN_TO_R,
      RETURN_TO_OLD_Z,
    } return_mode_t;

    static const int MAX_VAR = 26;

  protected:
    Machine &machine;

    double params[MAX_ADDRESS];
    typedef std::map<std::string, double> named_t;
    named_t named;

    double vars[MAX_VAR];
    cb::SmartPointer<Entity> varExprs[MAX_VAR];
    bool used[MAX_VAR];

    // Config variables
    double arcPrecision;
    double rapidFeed;
    std::string outputMoves;
    std::string scriptPath;

    cb::SmartPointer<std::ostream> moveStream;

    // State variables
    cb::LocationRange location;
    const Code *activeMotion;
    ToolTable tools;
    CoordinateSystem coordSystems[9];
    plane_t plane;
    bool latheDiameterMode;
    bool metric;
    bool cutterRadiusComp;
    bool toolLengthComp;
    path_mode_t pathMode;
    feed_mode_t feedMode;
    spindle_mode_t spindleMode;
    double maxSpindleSpeed;
    return_mode_t returnMode;
    double motionBlendingTolerance;
    double naiveCamTolerance;
    bool modalMotion;
    bool incrementalDistanceMode;
    bool arcIncrementalDistanceMode;
    dir_t spindleDir;
    bool moveInAbsoluteCoords;

    double feed;
    double speed;

  public:
    Controller(cb::Options &options, Machine &machine,
               double arcPrecision = 360);
    virtual ~Controller() {}

    void init();

    // State variables
    ToolTable &getToolTable() {return tools;}
    Tool &getTool(unsigned tool) {return *tools.get(tool);}
    const Tool &getTool(unsigned tool) const {return *tools.get(tool);}
    unsigned getCurrentTool() const {return (unsigned)get(TOOL_NUMBER);}
    const cb::LocationRange &getLocation() const {return location;}
    void setLocation(const cb::LocationRange &location)
    {this->location = location;}
    const Code *getActiveMotion() const {return activeMotion;}
    void setActiveMotion(const Code *activeMotion)
    {this->activeMotion = activeMotion;}
    CoordinateSystem &getCoordinateSystem(unsigned i);
    const CoordinateSystem &getCoordinateSystem(unsigned i) const;
    plane_t getPlane() const {return plane;}
    void setPlane(plane_t plane);
    bool getLatheDiameterMode() const {return latheDiameterMode;}
    bool getLatheRadiusMode() const {return !latheDiameterMode;}
    void setMetric(bool metric) {this->metric = metric;}
    bool getMetric() const {return metric;}
    bool getCutterRadiusComp() const {return cutterRadiusComp;}
    bool getToolLengthComp() const {return toolLengthComp;}
    path_mode_t getPathMode() const {return pathMode;}
    double getMotionBlendingTolerance() const {return motionBlendingTolerance;}
    double getNaiveCamTolerance() const {return naiveCamTolerance;}
    bool getModalMotion() const {return modalMotion;}
    bool getAbsoluteDistanceMode() const {return !incrementalDistanceMode;}
    bool getIncrementalDistanceMode() const {return incrementalDistanceMode;}
    dir_t getSpindleDir() const {return spindleDir;}
    void setSpindleDir(dir_t dir);
    void setMistCoolant(bool enable);
    bool getMistCoolant() const;
    void setFloodCoolant(bool enable);
    bool getFloodCoolant() const;

    // Parameters
    double get(unsigned addr) const
    {return addr < MAX_ADDRESS ? params[addr] : params[0];}
    void set(unsigned addr, double value)
    {if (addr < MAX_ADDRESS) params[addr] = value;}
    double get(const std::string &name) const;
    void set(const std::string &name, double value);

    // Variables
    void setVar(char c, double value)
    {vars[c - 'A'] = value; used[c - 'A'] = true;}
    double getVar(char c) const {return vars[c - 'A'];}
    void setVarExpr(char c, const cb::SmartPointer<Entity> &entity)
    {varExprs[c - 'A'] = entity;}
    const cb::SmartPointer<Entity> &getVarExpr(char c) const
    {return varExprs[c - 'A'];}
    double getOffsetVar(char c) const;

    static unsigned letterToVarType(char axis);
    static const char *getPlaneAxes(plane_t plane);
    static const char *getPlaneOffsets(plane_t plane);

    char getPlaneXAxis() const {return getPlaneAxes(getPlane())[0];}
    char getPlaneYAxis() const {return getPlaneAxes(getPlane())[1];}
    char getPlaneZAxis() const {return getPlaneAxes(getPlane())[2];}
    unsigned getPlaneXVarType() const {return letterToVarType(getPlaneXAxis());}
    unsigned getPlaneYVarType() const {return letterToVarType(getPlaneYAxis());}
    unsigned getPlaneZVarType() const {return letterToVarType(getPlaneZAxis());}

    double getFeed() const {return feed;}
    void setFeed(double feed);
    double getSpeed() const {return speed;}
    void setSpeed(double speed);
    void setTool(unsigned tool);

    // Position
    double getAxisAbsolutePosition(char axis) const;
    void setAxisAbsolutePosition(char axis, double pos);
    double getAxisOffset(char axis) const;
    double getAxisPosition(char axis) const;
    Axes getPosition(bool absolute = false) const;
    void setAxisPosition(char axis, double pos);
    void setPosition(const Axes &axes, bool absolute = false);
    Axes getNextPosition(int vars) const;

    const std::string &getScriptPath() const {return scriptPath;}

    std::string getVarGroupStr(const char *group, bool usedOnly = true) const;

    virtual void reset();
    virtual void newBlock();
    virtual void execute(const Code &code, int vars);
    virtual void doMove(MoveType type, const Axes &pos);
    virtual void makeMove(MoveType type, int vars);
    virtual void moveAxis(MoveType type, char axis, double value);

    virtual void arc(int vars, bool clockwise);
    virtual void dwell(double seconds);
    virtual void spindleSynchronizedMotion(bool rigidTap);
    virtual void straightProbe(int vars, bool towardWorkpiece,
                               bool signalError);
    virtual void drill(int vars, bool dwell, bool feedOut, bool spindleStop);
    virtual void setToolTable(int vars, bool relative);
    virtual void setCoordSystemOffset(unsigned cs, char axis, bool relative);
    virtual void setCoordSystem(int vars, bool relative);
    virtual void toolChange(bool manual = false);
    virtual void loadToolOffsets(unsigned tool);
    virtual void loadToolVarOffsets(int vars);
    virtual void storePredefined1();
    virtual void storePredefined2();
    virtual void loadPredefined1();
    virtual void loadPredefined2();
    virtual void setGlobalOffsets(int vars);
    virtual void resetGlobalOffsets(bool clearMemory);
    virtual void restoreGlobalOffsets();
    virtual void setCutterRadiusComp(int vars, bool left, bool dynamic);
    virtual void end();
  };
}

#endif // OPENSCAM_CONTROLLER_H

