/******************************************************************************\

    OpenSCAM is an Open-Source CAM software.
    Copyright (C) 2011-2012 Joseph Coffland <joseph@cauldrondevelopment.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

\******************************************************************************/

#include "GlutApp.h"

#include <openscam/view/GL.h>
#include <openscam/view/ViewPort.h>

#include <cbang/Exception.h>
#include <cbang/log/Logger.h>
#include <cbang/util/DefaultCatch.h>

#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

using namespace std;
using namespace cb;
using namespace OpenSCAM;


// GLUT callbacks
static void mouseCallback(int button, int state, int x, int y) try {
  GlutApp::instance().mouse(button, state, x, y);
} CBANG_CATCH_ERROR;


static void motionCallback(int x, int y) try {
  GlutApp::instance().motion(x, y);
} CBANG_CATCH_ERROR;


static void keysCallback(unsigned char key, int x, int y) try {
  GlutApp::instance().keys(key, x, y);
} CBANG_CATCH_ERROR;


static void specialCallback(int key, int x, int y) try {
  GlutApp::instance().special(key, x, y);
} CBANG_CATCH_ERROR;


static void renderCallback() try {
  GlutApp::instance().render();
} CBANG_CATCH_ERROR;


static void resizeCallback(int w, int h) try {
  GlutApp::instance().resize(w, h);
} CBANG_CATCH_ERROR;


static void visibilityCallback(int state) try {
  GlutApp::instance().visibility(state);
} CBANG_CATCH_ERROR;


static void idleCallback() try {
  GlutApp::instance().idle();
} CBANG_CATCH_ERROR;


GlutApp *GlutApp::singleton = 0;


GlutApp::GlutApp(const std::string &name) :
  Application(name), viewPort(new ViewPort(options)), fullscreen(false),
  visible(false), dirty(true), formerWidth(0), formerHeight(0),
  mouseLeftDragging(false), mouseMiddleDragging(false),
  mouseRightDragging(false) {

  if (singleton) THROW("There can be only on instance of GlutApp");
  singleton = this;

  options.pushCategory("View");
  options.add("fullscreen", "Display in fullscreen mode")->setDefault(false);
  options.popCategory();
}


GlutApp::~GlutApp() {}


GlutApp &GlutApp::instance() {
  if (!singleton) THROW("No instance of GlutApp");
  return *singleton;
}


int GlutApp::init(int argc, char *argv[]) {
  // Init GLUT
  glutInit(&argc, argv);

  // Init GL
  glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA);
  glutInitWindowPosition(0, 0);
  glutInitWindowSize(viewPort->getWidth(), viewPort->getHeight());
  // NOTE: OSX crashes if this is not in main()
  glutCreateWindow(getName().c_str());

  // Parse command line, etc.
  Application::init(argc, argv);

  GLenum err = glewInit();
  if (err != GLEW_OK) THROWS("Initializing GLEW: " << glewGetErrorString(err));

  // Init
  viewPort->init();
  viewPort->glInit();

  // Fullscreen (must be after GL init)
  if (options["fullscreen"].toBoolean()) toggleFullscreen();
  else glutReshapeWindow(viewPort->getWidth(), viewPort->getHeight());

  // GLUT callbacks
  glutMouseFunc(mouseCallback);
  glutMotionFunc(motionCallback);
  glutPassiveMotionFunc(motionCallback);
  glutKeyboardFunc(keysCallback);
  glutSpecialFunc(specialCallback);
  glutDisplayFunc(renderCallback);
  glutReshapeFunc(resizeCallback);
  glutVisibilityFunc(visibilityCallback);
  glutIdleFunc(idleCallback);

  return 0;
}


void GlutApp::run() {
  glutMainLoop();
}


void GlutApp::toggleFullscreen() {
  if (fullscreen) {
    glutReshapeWindow(formerWidth, formerHeight);

#ifdef _WIN32
    // Windows needs this but it shifts the window on Linux
    glutPositionWindow(formerX, formerY);
#endif

  } else {
    formerX = glutGet(GLUT_WINDOW_X);
    formerY = glutGet(GLUT_WINDOW_Y);
    formerWidth = viewPort->getWidth();
    formerHeight = viewPort->getHeight();
    glutFullScreen();
  }

  fullscreen = !fullscreen;
  redisplay();
}


void GlutApp::quit() {
  glutDestroyWindow(glutGetWindow());
  visible = false;

  exit(0);
}


void GlutApp::mouse(int button, int state, int x, int y) {
  switch (button) {
  case GLUT_LEFT_BUTTON:
    switch (state) {
    case GLUT_DOWN:
      viewPort->startRotation(x, y);
      mouseLeftDragging = true;
      break;

    case GLUT_UP: mouseLeftDragging = false; break;
    }
    break;

  case GLUT_MIDDLE_BUTTON:
    switch (state) {
    case GLUT_DOWN:
      viewPort->startTranslation(x, y);
      mouseMiddleDragging = true;
      break;

    case GLUT_UP: mouseMiddleDragging = false; break;
    }
    break;

  case GLUT_RIGHT_BUTTON:
    switch (state) {
    case GLUT_DOWN:
      mouseRightDragging = true;
      break;

    case GLUT_UP: mouseRightDragging = false; break;
    }
    break;

  case 3: viewPort->zoomIn(); break; // Wheel up
  case 4: viewPort->zoomOut(); break; // Wheel down
  }

  redisplay();
}


void GlutApp::motion(int x, int y) {
  mousePosition = Vector2I(x, y);

  // Ignore if mouse is out of our window
  if (x <= 0 || (int)viewPort->getWidth() <= x || y <= 0 ||
      (int)viewPort->getHeight() <= y)
    return;

  if (mouseLeftDragging) {
    viewPort->updateRotation(x, y);
    redisplay();

  } else if (mouseMiddleDragging) {
    viewPort->updateTranslation(x, y);
    redisplay();
  }
}


void GlutApp::keys(unsigned char key, int x, int y) {
  LOG_DEBUG(3, "Key pressed '" << key << "' == 0x" << hex << (unsigned)key);

  switch (key) {
  case 0x1b: case 'q': case 'Q': quit(); break;
  case 'f': toggleFullscreen(); break;
  case 'a': viewPort->toggleShowAxes(); break;
  case 'o': viewPort->toggleShowBounds(); break;
  case 'c': viewPort->center(); break;
  case 'p': case 't': case 'b': case 'F': case 'B': case 'l': case 'r':
    viewPort->resetView(key);
    break;
  default: break;
  }

  redisplay();
}


void GlutApp::special(int key, int x, int y) {
  LOG_DEBUG(3, "Special key pressed '" << key << "' == 0x"
            << hex << (unsigned)key);

  switch (key) {
  case GLUT_KEY_PAGE_UP: viewPort->zoomOut(); break;
  case GLUT_KEY_PAGE_DOWN: viewPort->zoomIn(); break;
  }

  redisplay();
}


void GlutApp::render() {
  if (!visible) return;

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  draw();
  glutSwapBuffers();

  renderTimer.throttle(1.0 / 60); // No faster than X Hz
}


void GlutApp::resize(unsigned width, unsigned height) {
  viewPort->resize(width, height);
}


void GlutApp::visibility(int state) {
  visible = state == GLUT_VISIBLE;
}


void GlutApp::idle() {
  if (shouldQuit()) quit();

  // Redraw
  if (dirty && visible) {
    dirty = false;
    glutPostRedisplay();
  }

  // Throttle
  idleTimer.throttle(0.01); // No faster than 100Hz
}
