/******************************************************************************\

    OpenSCAM is an Open-Source CAM software.
    Copyright (C) 2011-2012 Joseph Coffland <joseph@cauldrondevelopment.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

\******************************************************************************/

#include "SurfaceTestApp.h"

#include <cbang/log/Logger.h>
#include <cbang/time/Timer.h>

#include <openscam/view/ViewPort.h>
#include <openscam/view/GL.h>

#include <limits>

#if defined(_WIN32) && defined(max)
#undef max
#endif

using namespace std;
using namespace cb;
using namespace OpenSCAM;


SurfaceTestApp::SurfaceTestApp() :
  GlutApp("Surface Test"), height(60), radius(25), rotateTime(Timer::now()),
  testPt(100, 0, 0) {
}


// Parallelogram contains point
bool contains(const Vector3R &p1, const Vector3R &p2, const Vector3R &p3,
              const Vector3R &p) {
  // Compute P12, the vector normal to V1 = P2-P1 and V2 = P3-P1
  Vector3R p12 = (p2 - p1).cross(p3 - p1);

  // If the component of V = P-P1 in the P12 direction is too large,
  // then it does not lie in the parallelogram
  real dot = (p - p1).dot(p12);
  real v = p.distance(p2);
  // NOTE: Not 100% sure this is correct, original code had errors
  if (0.00001 + v < abs(dot)) return false;

  // Compute P23, the vector normal to V2 and V3
  Vector3R p23 = (p3 - p1).cross(p12);

  // Compute ALPHA = (V dot P23) / (V1 dot P23)
  real dott = (p - p1).dot(p23);
  real dotb = (p2 - p1).dot(p23);

  if (dotb < 0) {
    dott = -dott;
    dotb = -dotb;
  }

  if (dott < 0 || dotb < dott) return false;

  // Compute V31, the vector normal to P12 and V1
  Vector3R p31 = p12.cross(p2 - p1);

  // Compute BETA = (V dot V31) / (V2 dot V31)
  dott = (p - p1).dot(p31);
  dotb = (p3 - p1).dot(p31);

  if (dotb < 0) {
    dott = -dott;
    dotb = -dotb;
  }

  if (dott < 0 || dotb < dott) return false;

  // V = ALPHA * V1 + BETA * V2, where both ALPHA and BETA are between 0 and 1.
  return true;
}


// Closest point to Parallelogram
Vector3R closest(const Vector3R &start, const Vector3R &end,
                 real radius, real height, const Vector3R &p) {
  const Vector3R &p1 = start;
  const Vector3R p2 = p1 + Vector3R(0, 0, height);
  const Vector3R &p3 = end;
  const Vector3R p4 = p3 + Vector3R(0, 0, height);

  // Compute plane normal
  Vector3R n = (p2 - p1).cross(p3 - p1).normalize();

  // Project p on to plane
  real sn = -n.dot(p - p1);
  real sd = n.dot(n);
  real sb = sn / sd;
  Vector3R q = p + n * sb;

  Vector3R &closest = q;
  real dist;

  if (!contains(p1, p2, p3, q)) {
    // Find closest point on one of the line segments
    real dists[4];
    Vector3R points[4] = {
      Segment3R(p1, p3).closest(p, dists[0]), // Bottom
      Segment3R(p2, p4).closest(p, dists[1]), // Top
      Segment3R(p1, p2).closest(p, dists[2]), // Side
      Segment3R(p3, p4).closest(p, dists[3]), // Side
    };

    int min = 0;
    for (int i = 1; i < 4; i++) if (dists[i] < dists[min]) min = i;
    closest = points[min];
    dist = dists[min];

    if (2 <= min) return closest + (p - closest) / dist * radius; // Side
    else { // Top or bottom
      return closest;
    }
  }

  Vector3R l = n * radius;
  l = 0 < sb ? (closest - l) : (closest + l);
  closest = l;

  return closest;
}


int SurfaceTestApp::init(int argc, char *argv[]) {
  int ret = GlutApp::init(argc, argv);

  cylinder = new Cylinder(radius, height);

  return ret;
}


void SurfaceTestApp::draw() {
  Rectangle3R bbox;
  bbox.add(Vector3R(-100, -100, -100));
  bbox.add(Vector3R(100, 100, 100));

  getViewPort()->glDraw(bbox);

  getViewPort()->setLighting(true);

  const int resolution = 200;
  Vector3R start(0, 0, 0);
  Vector3R end(40, -40, -100);
  Vector3R tran = (end - start) / (real)resolution;

  Vector3R p = testPt + Vector3R(0, 0, 200);
  double angle = 2 * M_PI * fmod(Timer::now() - rotateTime, 30.0) / 30.0;
  p = AxisAngle(angle, 0, 0, 1).rotate(p);

  for (int j = 0; j < 400; j++) {
    Vector3R minPt;
    real minDist = numeric_limits<real>::max();

    // Cylinders
    cylinder->setPosition(start);
    for (int i = 0; i < resolution; i++) {
      Vector3R q = cylinder->closest(p);
      real d = q.distanceSquared(p);
      if (d < minDist) {
        minDist = d;
        minPt = q;
      }

      if (!j) {
#if 0
        // Line
        glColor4f(1, 0, 0, 1); // Red
        glLineWidth(1);
        glBegin(GL_LINES);
        glVertex3f(p.x(), p.y(), p.z());
        glVertex3f(q.x(), q.y(), q.z());
        glEnd();
#endif

        glColor4f(1, 0.5, 0, 0.8); // Orange
        //cylinder->draw();
      }

      cylinder->translate(tran);
    }

    // Closest Line
    glLineWidth(1);
#if 0
    glColor4f(0, 1, 0, 1); // Green
    glBegin(GL_LINES);
    glVertex3f(p.x(), p.y(), p.z());
    glVertex3f(minPt.x(), minPt.y(), minPt.z());
    //Vector3R l = p + (minPt - p) * 4;
    //glVertex3f(l.x(), l.y(), l.z());
    glEnd();
#endif

    Vector3R c = closest(start, end, radius, height, p);
    glColor4f(1, 0, 0, 1); // Red
    glBegin(GL_LINES);
    glVertex3f(p.x(), p.y(), p.z());
    glVertex3f(c.x(), c.y(), c.z());
    glEnd();

    p.z() -= 1;
  }

  // Plane
  glColor4f(0.25, 0.25, 0.25, 1); // Grey
  glBegin(GL_QUADS);
  glVertex3f(start.x(), start.y(), start.z());
  glVertex3f(end.x(), end.y(), end.z());
  glVertex3f(end.x(), end.y(), end.z() + height);
  glVertex3f(start.x(), start.y(), start.z() + height);
  glEnd();

  getViewPort()->setLighting(false);

  redisplay();
}


void SurfaceTestApp::keys(unsigned char key, int x, int y) {
  switch (key) {
  case '8': testPt.y() += 1; break;
  case '2': testPt.y() -= 1; break;
  case '6': testPt.x() += 1; break;
  case '4': testPt.x() -= 1; break;
  case '9': testPt.z() += 1; break;
  case '1': testPt.z() -= 1; break;
  }

  GlutApp::keys(key, x, y);
}
