#include "Environment.h"

#include "ast/FunctionDecl.h"
#include "ast/VariableDecl.h"
#include "ast/ConstantDecl.h"
#include "ast/ImportDecl.h"
#include "ast/Program.h"

#include <cbang/log/Logger.h>

using namespace std;
using namespace cb;
using namespace OpenSCAM::TPL;


void Environment::setParent(Environment *parent) {
  if (this->parent) THROW("Parent already set");
  this->parent = parent;
  id = parent->getNextID();
}


void Environment::declare(const SmartPointer<Module> &module,
                          const string &as) {
  if (as == ".") {
    env_t::const_iterator it;
    Environment &modEnv = *module->getEnvironment();
    for (it = modEnv.env.begin(); it != modEnv.env.end(); it++) {
      pair<env_t::iterator, bool> result = env.insert(*it);
      if (!result.second) result.first->second = it->second;
    }

  } else declare(as.empty() ? module->getName() : as, module);
}


void Environment::declare(const string &name,
                          const SmartPointer<Declaration> &decl) {
  pair<env_t::iterator, bool> result =
    env.insert(env_t::value_type(name, decl));

  if (!result.second)
    THROWS(name << " already declared as "
           << result.first->second->getType() << " at "
           << result.first->second->getLocation());

  if (decl->getType() == TPLID_ARGUMENT) {
    decl->setIndex(nextIndex++);
    decl->setEnvID(id);
  }
}


SmartPointer<Declaration> Environment::declareVariable(const string &name) {
  env_t::iterator it = env.find(name);

  if (it == env.end()) {
    SmartPointer<Declaration> var = new VariableDecl(name);

    var->setIndex(nextIndex++);
    var->setEnvID(id);
    env.insert(env_t::value_type(name, var));

    return var;
  }

  const SmartPointer<Declaration> &decl = it->second;
  if (decl->getType() != TPLID_VARIABLE)
    THROWS("'" << name << "' already declared as " << decl->getType());

  return decl;
}


void Environment::declareConstant(const string &name,
                                  const SmartPointer<Value> &value) {
  env_t::iterator it = env.find(name);

  if (it != env.end())
    THROWS("'" << name << "' already declared as " << it->second->getType());

  SmartPointer<Declaration> var = new ConstantDecl(name, value);
  env.insert(env_t::value_type(name, var));
}


void Environment::declareBoolConstant(const string &name, bool value) {
  declareConstant(name, new BoolValue(value));
}


void Environment::declareIntegerConstant(const string &name, int32_t value) {
  declareConstant(name, new IntegerValue(value));
}


void Environment::declareFloatConstant(const string &name, double value) {
  declareConstant(name, new FloatValue(value));
}


void Environment::declareStringConstant(const string &name,
                                        const string &value) {
  declareConstant(name, new StringValue(value));
}


const SmartPointer<Declaration> &Environment::lookup(const Name &name) const {
  const Environment *env = this;

  string s;
  for (unsigned i = 0; i < name.getName().size(); i++) {
    const SmartPointer<Declaration> &decl = env->lookup(name.getName()[i]);

    if (i + 1 == name.getName().size()) return decl;

    if (decl->getType() != TPLID_MODULE) {
      string s;
      for (unsigned j = 0; j <= i; j++) {
        if (!s.empty()) s += ".";
        s += name.getName()[j];
      }

      THROWS("'" << s << "' is a " << decl->getType() << " not an namespace");
    }

    if (!dynamic_cast<Module *>(decl.get())->isLoaded())
      THROWS("Module '" << name.getName()[i] << "' not loaded");

    env = decl->getEnvironment().get();
  }

  THROW("Internal error"); // Cannot get here
}


const SmartPointer<Declaration> &Environment::lookup(const string &name) const {
  env_t::const_iterator it = env.find(name);
  if (it == env.end()) {
    if (parent) return parent->lookup(name);
    THROWS("'" << name << "' is undefined");
  }

  return it->second;
}


unsigned Environment::getNextID() {
  return ++children;
}
