#include "Evaluator.h"

#include "Context.h"
#include "type/Types.h"

#include "ast/AST.h"

#include <cbang/String.h>
#include <cbang/Math.h>

#include <map>

using namespace std;
using namespace cb;
using namespace OpenSCAM::TPL;


SmartPointer<Value> Evaluator::eval(TPLID op,
                                    const SmartPointer<Value> &left,
                                    const SmartPointer<Value> &right) {
  switch (op) {
  case TPLID_ADD:
    // String concatenation
    if (left->isString() && right->isString())
      return new StringValue(left->toString() + right->toString());

  case TPLID_SUB:
  case TPLID_DIV:
  case TPLID_MUL:
  case TPLID_MOD:
    // Vector/Vector operations
    if (left->isVector() && right->isVector()) {
      const vector<SmartPointer<Value> > &leftVect = left->toVector();
      const vector<SmartPointer<Value> > &rightVect = right->toVector();

      if (leftVect.size() != rightVect.size())
        THROWS("Cannot perform binary operation on vectors of different size");

      SmartPointer<VectorValue> vector = new VectorValue;

      for (unsigned i = 0; i < leftVect.size(); i++)
        vector->push_back(eval(op, leftVect[i], rightVect[i]));

      return vector;
    }

    // Vector/Number operations
    if (left->isVector()) {
      if (!right->isNumber())
        THROWS("Cannot perform binary operation on vector and "
               << right->getType());

      const vector<SmartPointer<Value> > &leftVect = left->toVector();
      SmartPointer<VectorValue> vector = new VectorValue;

      for (unsigned i = 0; i < leftVect.size(); i++)
        vector->push_back(eval(op, leftVect[i], right));

      return vector;
    }

    // Number/Vector operations
    if (right->isVector()) {
      if (!left->isNumber())
        THROWS("Cannot perform binary operation on " << right->getType()
               << " and vector");

      const vector<SmartPointer<Value> > &rightVect = right->toVector();
      SmartPointer<VectorValue> vector = new VectorValue;

      for (unsigned i = 0; i < rightVect.size(); i++)
        vector->push_back(eval(op, left, rightVect[i]));

      return vector;
    }

    // Number/Number operations
    if (!left->isNumber() || !right->isNumber())
      THROWS("Cannot apply operator " << op << " to "
             << left->getType() << " and " << right->getType());

    if (left->isInteger() && right->isInteger())
      switch (op) {
      case TPLID_ADD:
        return new IntegerValue(left->toInteger() + right->toInteger());
      case TPLID_SUB:
        return new IntegerValue(left->toInteger() - right->toInteger());
      case TPLID_DIV:
        return new IntegerValue(left->toInteger() / right->toInteger());
      case TPLID_MUL:
        return new IntegerValue(left->toInteger() * right->toInteger());
      case TPLID_MOD:
        return new IntegerValue(left->toInteger() % right->toInteger());
      default: THROWS("Internal error");
      }
    else switch (op) {
      case TPLID_ADD:
        return new FloatValue(left->toFloat() + right->toFloat());
      case TPLID_SUB:
        return new FloatValue(left->toFloat() - right->toFloat());
      case TPLID_DIV:
        return new FloatValue(left->toFloat() / right->toFloat());
      case TPLID_MUL:
        return new FloatValue(left->toFloat() * right->toFloat());
      case TPLID_MOD:
        return new FloatValue(fmod(left->toFloat(), right->toFloat()));
      default: THROWS("Internal error");
      }

  case TPLID_LESS: return new BoolValue(left->less(*right));
  case TPLID_LESS_EQUAL:
    return new BoolValue(left->less(*right) || left->equal(*right));
  case TPLID_GREATER:
    return new BoolValue(!(left->less(*right) || left->equal(*right)));
  case TPLID_GREATER_EQUAL: return new BoolValue(!left->less(*right));
  case TPLID_EQUAL: return new BoolValue(left->equal(*right));
  case TPLID_NOT_EQUAL: return new BoolValue(!left->equal(*right));

  default: THROWS("Internal error");
  }
}


SmartPointer<Value> Evaluator::eval(BinaryExpr &node) {
  SmartPointer<Value> left = node.getLeft()->eval(*this);
  SmartPointer<Value> right;

  if (node.getType() != TPLID_AND && node.getType() != TPLID_OR) {
    right = node.getRight()->eval(*this);
    return eval(node.getType(), left, right);
  }

  // Special case to short-circuit && and ||
  if (node.getType() == TPLID_AND)
    return new BoolValue(left->toBool() &&
                         node.getRight()->eval(*this)->toBool());

  return new BoolValue(left->toBool() ||
                       node.getRight()->eval(*this)->toBool());
}


SmartPointer<Value> Evaluator::eval(CallExpr &node) {
  // NOTE: Compatibility between the call and function declaration has already
  //   been check thoroughly at parse time so no error checking is done here.
  FunctionDecl &func =
    *dynamic_cast<FunctionDecl *>(node.getDeclaration().get());

  // Allocate args
  Arguments args(func.size() < node.size() ? node.size() : func.size());

  // Evaluate expressions & assign positional arguments
  for (unsigned i = 0; i < node.size(); i++) {
    const string &name = node[i].first;
    SmartPointer<Value> value = node[i].second->eval(*this);

    if (name.empty()) args[i] = value;
    else args[func.getEnvironment()->lookup(name)->getIndex()] = value;
  }

  // Fill in default values
  for (unsigned i = 0; i < func.size(); i++)
    if (args[i].isNull()) args[i] = func[i]->getDefault()->eval(*this);

  // Make the call
  SmartPointer<Value> result = func.call(ctx, args);
  return result.isNull() ? NullValue::instance() : result;
}


SmartPointer<Value> Evaluator::eval(QuotedExpr &node) {
  return node.getExpression()->eval(*this);
}


SmartPointer<Value> Evaluator::eval(SliceExpr &node) {
  SmartPointer<Value> value = node.getExpression()->eval(*this);

  // Vector
  if (!value->isVector()) THROW("Cannot perform slice on non-vector");

  const VectorValue &vector = *dynamic_cast<VectorValue *>(value.get());
  int length = (int)vector.size();

  // Start Expression
  SmartPointer<Value> start = node.getStart()->eval(*this);

  if (!start->isInteger())
    THROW("Slice start expression did not evaluate to an integer");

  int32_t startValue = start->toInteger();
  if (startValue < 0) startValue += length;
  if (startValue < 0 || length <= startValue)
    THROWS("Slice start expression value " << startValue
           << " is out of range for vector with length " << length);

  if (node.getEnd().isNull()) return vector[startValue];
  else {
    // End Expression
    SmartPointer<Value> end =
      node.getEnd().isNull() ? 0 : node.getEnd()->eval(*this);

    if (!end.isNull() && !end->isInteger())
      THROW("Slice end expression did not evaluate to an integer");

    int32_t endValue = end.isNull() ? length : end->toInteger();
    if (endValue < 0) endValue += length;
    if (endValue < 0 || length < endValue)
      THROWS("Slice end expression value " << endValue
             << " is out of range for vector with length " << length);

    // Result
    SmartPointer<VectorValue> result = new VectorValue;
    for (int i = startValue; i < endValue; i++)
      result->push_back(vector[i]);

    return result;
  }
}


SmartPointer<Value> Evaluator::eval(UnaryExpr &node) {
  SmartPointer<Value> value = node.getExpression()->eval(*this);

  if (node.getType() == TPLID_SUB) {
    if (value->isInteger()) return new IntegerValue(-value->toInteger());
    else if (value->isFloat()) return new FloatValue(-value->toInteger());
    else THROW("Cannot negate non-number");

  } else if (node.getType() == TPLID_NOT)
    return new BoolValue(!value->toBool());

  THROW("Internal error");
}


SmartPointer<Value> Evaluator::eval(VariableExpr &node) {
  Declaration &decl = *node.getDeclaration();
  if (decl.getType() == TPLID_CONSTANT) return decl.getValue();

  SmartPointer<Value> &value = ctx.getReference(decl);
  if (value.isNull()) return NullValue::instance();
  return value;
}


SmartPointer<Value> Evaluator::eval(VectorExpr &node) {
  SmartPointer<VectorValue> vector = new VectorValue;

  for (unsigned i = 0; i < node.size(); i++)
    vector->push_back(node[i]->eval(*this));

  return vector;
}


SmartPointer<Value> Evaluator::eval(Constant &node) {
  return node.getValue();
}
