#include "Interpreter.h"

#include "Parser.h"
#include "Context.h"

#include "type/Types.h"
#include "SmartStackFrame.h"

#include "ast/AST.h"

#include <cbang/Exception.h>

using namespace std;
using namespace cb;
using namespace OpenSCAM::TPL;


void Interpreter::interpret(const cb::InputSource &source) {
  SmartPointer<BlockStmt> block = Parser(ctx).parse(source);
  SmartStackFrame smartStackFrame(ctx, *ctx.getEnvironment());
  return interpret(*block);
}


void Interpreter::interpret(Program &program) {
  interpret((BlockStmt &)program);
}


void Interpreter::interpret(ImportDecl &import) {
  import.init(*this);
}


void Interpreter::interpret(ConstantDecl &constant) {
  constant.setValue(constant.getExpression()->eval(evaluator));
}


void Interpreter::interpret(AssignStmt &stmt) {
  // Variable
  Declaration &decl = *stmt.getVariable()->getDeclaration();
  cb::SmartPointer<Value> *ref = &ctx.getReference(decl);

  const AssignStmt::indices_t &indices = stmt.getIndices();
  for (unsigned i = 0; i < indices.size(); i++) {
    Value &value = **ref;
    LocationRange &location = indices[i]->getLocation();

    if (!value.isVector())
      THROWS("Cannot index non-vector " << value.getType() << " at "
             << location);

    VectorValue &vector = dynamic_cast<VectorValue &>(value);

    SmartPointer<Value> indexValue = indices[i]->eval(evaluator);
    if (!indexValue->isInteger())
      THROWS("Cannot index vector with non-integer " << indexValue->getType()
             << " at " << location);

    unsigned index = indexValue->toInteger();

    if (vector.size() <= index)
      THROWS("Index " << index << " out of range for vector of size "
             << vector.size() << " at " << location);

    ref = &vector[index];
  }

  // Expression
  SmartPointer<Value> value = stmt.getExpression()->eval(evaluator);

  if (stmt.getType() != TPLID_ASSIGN) {
    TPLID op;
    switch (stmt.getType()) {
    case TPLID_ADD_ASSIGN: op = TPLID_ADD; break;
    case TPLID_SUB_ASSIGN: op = TPLID_SUB; break;
    case TPLID_DIV_ASSIGN: op = TPLID_DIV; break;
    case TPLID_MUL_ASSIGN: op = TPLID_MUL; break;
    case TPLID_MOD_ASSIGN: op = TPLID_MOD; break;
    default: THROW("Internal error");
    }

    value = evaluator.eval(op, *ref, value);
  }

  // Assign
  *ref = value;
}


void Interpreter::interpret(BlockStmt &block) {
  for (unsigned i = 0; i < block.size(); i++)
    block[i]->interpret(*this);
}


void Interpreter::interpret(BreakStmt &stmt) {
  throw stmt;
}


void Interpreter::interpret(CaseStmt &stmt) {
  stmt.getStatement()->interpret(*this);
}


void Interpreter::interpret(ContinueStmt &stmt) {
  throw stmt;
}


void Interpreter::interpret(DoStmt &stmt) {
  Expression &expr = *stmt.getExpression();

  do {
    try {
      stmt.getStatement()->interpret(*this);
    } catch (ContinueStmt &) {
      continue;
    } catch (BreakStmt &) {
      break;
    }
  } while (expr.eval(evaluator)->toBool());
}


void Interpreter::interpret(ExpressionStmt &stmt) {
  stmt.getExpression()->eval(evaluator);
}


void Interpreter::interpret(ForStmt &stmt) {
  const SmartPointer<AssignStmt> &init = stmt.getInit();
  const SmartPointer<Expression> &cond = stmt.getCondition();
  const SmartPointer<AssignStmt> &inc = stmt.getIncrement();

  if (!init.isNull()) init->interpret(*this);

  while (cond.isNull() || cond->eval(evaluator)->toBool()) {
    try {
      stmt.getStatement()->interpret(*this);
    } catch (ContinueStmt &) {
      continue;
    } catch (BreakStmt &) {
      break;
    }

    if (!inc.isNull()) inc->interpret(*this);
  }
}


void Interpreter::interpret(IfStmt &stmt) {
  if (stmt.getExpression()->eval(evaluator)->toBool()) {
    if (!stmt.getIfStmt().isNull()) stmt.getIfStmt()->interpret(*this);
  } else if (!stmt.getElseStmt().isNull()) stmt.getElseStmt()->interpret(*this);
}


void Interpreter::interpret(ReturnStmt &stmt) {
  if (stmt.getExpression().isNull()) throw NullValue::instance();
  throw stmt.getExpression()->eval(evaluator);
}


void Interpreter::interpret(SwitchStmt &stmt) {
  SmartPointer<Value> value = stmt.getExpression()->eval(evaluator);

  // TODO this could me more efficient

  for (unsigned i = 0; i < stmt.size(); i++) {
    const CaseStmt::exprs_t &exprs = stmt[i]->getExpressions();
    unsigned j;
    for (j = 0; j < exprs.size(); j++)
      if (exprs[j]->eval(evaluator)->equal(*value)) break;
    if (j == exprs.size()) continue;

    // Match found
    stmt[i]->getStatement()->interpret(*this);
    return;
  }

  if (!stmt.getDefault().isNull()) stmt.getDefault()->interpret(*this);
}


void Interpreter::interpret(WhileStmt &stmt) {
  Expression &expr = *stmt.getExpression();

  while (expr.eval(evaluator)->toBool()) {
    try {
      stmt.getStatement()->interpret(*this);
    } catch (ContinueStmt &) {
      continue;
    } catch (BreakStmt &) {
      break;
    }
  }
}
