#include "Lexer.h"

#include <cbang/String.h>

using namespace std;
using namespace OpenSCAM::TPL;


Lexer::Lexer(cb::Scanner &scanner) : cb::Tokenizer<TPLID>(scanner) {
  for (unsigned i = FIRST_KEYWORD; i <= LAST_KEYWORD; i++) {
    string name =
      cb::String::toLower(TPLID::toString((TPLID::enum_t)i));
    keywords[name] = TPLID((TPLID::enum_t)i);
  }

  advance();
}


void Lexer::doID() {
  // TODO support unicode ids

  string value;
  char c = scanner.peek();
  while (isalpha(c) || isdigit(c) || c == '_') {
    value.append(string(1, c));
    scanner.advance();
    c = scanner.peek();
  }

  keywords_t::iterator it = keywords.find(value);
  if (it == keywords.end()) current.set(TPLID_ID, value);
  else current.set(it->second, value);
}


void Lexer::doNumber(bool dot) {
  string value;
  bool foundDot = dot;
  char c = scanner.peek();

  if (dot) value = ".";

  while (c != -1 && (isdigit(c) || (!foundDot && c == '.'))) {
    if (c == '.') foundDot = true;
    value.append(string(1, c));
    scanner.advance();
    c = scanner.peek();
  }

  if (foundDot && value.length() == 1) {
    current.set(TPLID_DOT, '.');
    return;
  }

  current.set(foundDot ? TPLID_FLOAT : TPLID_INTEGER, value);
}


static bool ishex(int c) {
  return
    ('0' <= c && c <= '9') || ('a' <= c && c <= 'f') || ('A' <= c && c <= 'F');
}


static int hex_nibble(int c) {
  if ('0' <= c && c <= '9') return c - '0';
  if ('a' <= c && c <= 'f') return c - 'a' + 10;
  if ('A' <= c && c <= 'F') return c - 'A' + 10;
  THROW("Not a hex character");
}


void Lexer::doString() {
  char quote = (char)scanner.peek();
  scanner.advance();

  string value;

  while (scanner.hasMore()) {
    bool needAdvance = true;
    int c = scanner.peek();

    if (c == quote) {
      scanner.advance();
      current.set(TPLID_STRING, value);
      return;
    }

    if (c == '\\') {
      scanner.advance();
      c = scanner.peek();
      if (c == -1) break;

      switch (c) {
      case 'a': c = 0x07; break;
      case 'b': c = 0x08; break;
      case 'f': c = 0x0c; break;
      case 'n': c = 0x0a; break;
      case 'r': c = 0x0d; break;
      case 't': c = 0x09; break;
      case 'v': c = 0x0b; break;

      case 'u':
      case 'U': {
        int type = c;
        c = 0;
        for (unsigned i = 0; i < (type == 'u' ? 4 : 8); i++) {
          scanner.advance();
          if (ishex(scanner.peek())) {
            c = (c << 4) + hex_nibble(scanner.peek());
          } else {needAdvance = false; break;}
        }

        // TODO support unicode escapes
        if (c) THROW("Unicode escape sequences not yet supported");
        break;
      }

      case 'x':
        needAdvance = false;
        scanner.advance();
        c = scanner.peek();
        if (ishex(c)) {
          c = hex_nibble(c);
          scanner.advance();

          if (ishex(scanner.peek())) {
            c = (c << 4) + hex_nibble(scanner.peek());
            scanner.advance();
          }
        } else c = 'x';
        break;

      case '0': case '1': case '2': case '3':
      case '4': case '5': case '6': case '7':
        needAdvance = false;
        c -= '0';
        scanner.advance();
        if ('0' <= scanner.peek() && scanner.peek() <= '7') {
          c = (c << 3) + scanner.peek() - '0';

          scanner.advance();
          if ('0' <= scanner.peek() && scanner.peek() <= '7') {
            c = (c << 3) + scanner.peek() - '0';
            if (255 < c) THROW("Invalid octal escape");
            scanner.advance();
          }
        }
        break;
      }
    }

    value.append(1, (char)c);
    if (needAdvance) scanner.advance();
  }

  THROW("Could not find end of string");
}


void Lexer::doLineComment() {
  // TODO Save comments and append them to the next token
  while (scanner.hasMore() && scanner.peek() != '\n') scanner.advance();
}


void Lexer::doComment() {
  // TODO Save comments and append them to the next token
  bool star = false;
  while (scanner.hasMore()) {
    if (star && scanner.peek() == '/') {
      scanner.advance();
      return;
    }
    star = scanner.peek() == '*';
    scanner.advance();
  }

  THROW("Could not find end of comment");
}


void Lexer::next() {
  scanner.skipWhiteSpace();

  cb::FileLocation start = scanner.getLocation();

  if (!scanner.hasMore()) {
    current.set(TPLID_EOF, "");
    return;
  }

  bool dot = false;
  char c = (char)scanner.peek();
  switch (c) {
  case '\'': case '"': doString(); break;

  case '.':
    scanner.advance();
    if (scanner.peek() == '.') {
      scanner.advance();
      if (scanner.peek() != '.') THROW("Unexpected '..'");
      current.set(TPLID_ETC, "...");
      break;
    }
    dot = true;

  case '0': case '1': case '2': case '3': case '4':
  case '5': case '6': case '7': case '8': case '9': doNumber(dot); break;

  case '(': current.set(TPLID_OPAREN, c); scanner.advance(); break;
  case ')': current.set(TPLID_CPAREN, c); scanner.advance(); break;
  case '[': current.set(TPLID_OBRACKET, c); scanner.advance(); break;
  case ']': current.set(TPLID_CBRACKET, c); scanner.advance(); break;
  case '{': current.set(TPLID_OBRACE, c); scanner.advance(); break;
  case '}': current.set(TPLID_CBRACE, c); scanner.advance(); break;

  case ',': current.set(TPLID_COMMA, c); scanner.advance(); break;
  case ':': current.set(TPLID_COLON, c); scanner.advance(); break;
  case ';': current.set(TPLID_SEMI, c); scanner.advance(); break;

  case '=':
    scanner.advance();
    if (scanner.peek() == '=') {current.set(TPLID_EQUAL, c); scanner.advance();}
    else current.set(TPLID_ASSIGN, c);
    break;

  case '+':
    scanner.advance();
    if (scanner.peek() == '=') {
      current.set(TPLID_ADD_ASSIGN, c);
      scanner.advance();
    } else current.set(TPLID_ADD, c);
    break;

  case '-':
    scanner.advance();
    if (scanner.peek() == '=') {
      current.set(TPLID_SUB_ASSIGN, c);
      scanner.advance();
    } else current.set(TPLID_SUB, c);
    break;

  case '/':
    scanner.advance();
    switch (scanner.peek()) {
    case '=': current.set(TPLID_DIV_ASSIGN, c); scanner.advance(); break;
    case '/': scanner.advance(); doLineComment(); next(); return;
    case '*': scanner.advance(); doComment(); next(); return;
    default: current.set(TPLID_DIV, c); break;
    }
    break;

  case '*':
    scanner.advance();
    if (scanner.peek() == '=') {
      current.set(TPLID_MUL_ASSIGN, c);
      scanner.advance();
    } else current.set(TPLID_MUL, c);
    break;

  case '%':
    scanner.advance();
    if (scanner.peek() == '=') {
      current.set(TPLID_MOD_ASSIGN, c);
      scanner.advance();
    } else current.set(TPLID_MOD, c);
    break;

  case '&':
    scanner.advance();
    if (scanner.peek() == '&') {
      current.set(TPLID_AND, c);
      scanner.advance();
    } else current.set(TPLID_AMPERSAND, c);
    break;

  case '|':
    scanner.advance();
    if (scanner.peek() == '|') {
      current.set(TPLID_OR, c);
      scanner.advance();
    } else current.set(TPLID_BAR, c);
    break;

  case '<':
    scanner.advance();
    if (scanner.peek() == '=') {
      current.set(TPLID_LESS_EQUAL, c);
      scanner.advance();
    } else current.set(TPLID_LESS, c);
    break;

  case '>':
    scanner.advance();
    if (scanner.peek() == '=') {
      current.set(TPLID_GREATER_EQUAL, c);
      scanner.advance();
    } else current.set(TPLID_GREATER, c);
    break;

  case '!':
    scanner.advance();
    if (scanner.peek() == '=') {
      current.set(TPLID_NOT_EQUAL, c);
      scanner.advance();
    } else current.set(TPLID_NOT, c);
    break;

  default:
    if (isalpha(c) || c == '_') doID();
    else THROWS("Unexpected character: '"
                << cb::String::escapeC(string(1, c)) << "'");
  }

  current.getLocation() = cb::LocationRange(start, scanner.getLocation());
}
