#include "Parser.h"

#include "Module.h"
#include "Loader.h"
#include "Context.h"

#include <openscam/tpl/type/Types.h>

#include <cbang/parse/ParseScope.h>

#include <cbang/os/SystemUtilities.h>
#include <cbang/String.h>
#include <cbang/Exception.h>

#include <set>

using namespace std;
using namespace cb;
using namespace OpenSCAM::TPL;


Parser::Parser(const SmartPointer<Environment> &env,
               const SmartPointer<Loader> &loader) : loader(loader) {
  envStack.push_back(env);
}


Parser::Parser(Context &ctx) : loader(ctx.getLoader()) {
  envStack.push_back(ctx.getEnvironment());
}


Parser::~Parser() {} // Hide member destructors


SmartPointer<BlockStmt> Parser::parse(Lexer &lexer) {
  try {
    return program(lexer);
  } catch (const Exception &e) {
    THROWCS("Exception @" << lexer.getLocation(), e);
  }
}


SmartPointer<BlockStmt> Parser::parse(const InputSource &source) {
  if (source.getName().empty()) cwd = SystemUtilities::getcwd();
  else cwd = SystemUtilities::getCanonicalPath
         (SystemUtilities::dirname(source.getName()));

  Scanner scanner(source);
  Lexer lexer(scanner);
  return parse(lexer);
}


SmartPointer<Program> Parser::program(Lexer &lexer) {
  ParseScope scope(lexer.getScanner());
  SmartPointer<Program> program = new Program;

  blockContents(lexer, *program);

  if (lexer.getType() != TPLID_EOF) THROWS("Expected end of file");

  return scope.set(program);
}


SmartPointer<ImportDecl> Parser::partialImportDecl(Lexer &lexer) {
  ParseScope scope(lexer.getScanner());

  lexer.match(TPLID_FROM);
  string n = name(lexer)->toString();

  lexer.match(TPLID_IMPORT);
  string id = lexer.match(TPLID_ID).getValue();
  if (lexer.getType() == TPLID_DOT)
    THROW("Cannot qualify id in partial import");

  string as;
  if (lexer.consume(TPLID_AS)) {
    as = lexer.match(TPLID_ID).getValue();
    if (lexer.getType() == TPLID_DOT)
      THROW("Cannot qualify name in import as");
  }

  SmartPointer<Module> module = load(n);
  SmartPointer<Declaration> decl = module->getEnvironment()->lookup(id);
  getEnv().declare(as.empty() ? id : as, decl);

  return scope.set(new ImportDecl(n, id, as, module));
}


SmartPointer<ImportDecl> Parser::importDecl(Lexer &lexer) {
  ParseScope scope(lexer.getScanner());

  lexer.match(TPLID_IMPORT);
  string n = name(lexer)->toString();

  string as;
  if (lexer.consume(TPLID_AS)) {
    if (lexer.getType() == TPLID_DOT) {
      as = ".";
      lexer.advance();

    } else {
      as = lexer.match(TPLID_ID).getValue();
      if (lexer.getType() == TPLID_DOT)
        THROW("Cannot qualify name in import as");
    }
  }

  SmartPointer<Module> module = load(n);
  getEnv().declare(module, as);

  return scope.set(new ImportDecl(n, "", as, module));
}


SmartPointer<ConstantDecl> Parser::constantDecl(Lexer &lexer) {
  ParseScope scope(lexer.getScanner());

  lexer.match(TPLID_CONSTANT);

  string name = lexer.match(TPLID_ID).getValue();
  if (lexer.getType() == TPLID_DOT)
    THROW("Cannot qualify name of constant in declaration");

  lexer.match(TPLID_ASSIGN);

  SmartPointer<Expression> expr = expression(lexer);
  if (!expr->isConstant())
    THROW("Expression in constant declaration is not constant");

  SmartPointer<ConstantDecl> decl = new ConstantDecl(name, expr);
  getEnv().declare(name, decl);

  return scope.set(decl);
}


SmartPointer<FunctionDecl> Parser::functionDecl(Lexer &lexer) {
  ParseScope scope(lexer.getScanner());

  lexer.match(TPLID_FUNCTION);

  string name = lexer.match(TPLID_ID).getValue();
  if (lexer.getType() == TPLID_DOT)
    THROW("Cannot qualify function name in declaration");

  lexer.match(TPLID_OPAREN);

  SmartPointer<FunctionDecl> function = new FunctionDecl(name);

  EnvScope envScope(envStack, function->getEnvironment());

  function->setVariableArgs(arguments(lexer, *function));

  lexer.match(TPLID_CPAREN);

  getEnv().getParent()->declare(name, function);

  function->setBlock(blockStmt(lexer));

  return scope.set(function);
}


bool Parser::arguments(Lexer &lexer, ArgumentDecls &args) {
  set<string> used;
  bool variableArgs = false;

  if (lexer.getType() == TPLID_ID || lexer.getType() == TPLID_ETC)
    while (true) {
      if (lexer.getType() == TPLID_ETC) {
        variableArgs = true;
        lexer.advance();
        break;
      }

      args.push_back(argument(lexer));

      const string &arg = args.back()->getName();
      if (!used.insert(arg).second)
        THROWS("multiple arguments named '" << arg << "'");

      if (!lexer.consume(TPLID_COMMA)) break;
    }

  return variableArgs;
}


SmartPointer<ArgumentDecl> Parser::argument(Lexer &lexer) {
  ParseScope scope(lexer.getScanner());

  string name = lexer.match(TPLID_ID).getValue();
  SmartPointer<ArgumentDecl> argument = new ArgumentDecl(name);

  if (lexer.getType() == TPLID_DOT)
    THROW("Cannot qualify function argument name");

  getEnv().declare(name, argument);

  if (lexer.consume(TPLID_ASSIGN)) argument->setDefault(expression(lexer));

  return scope.set(argument);
}


SmartPointer<Statement> Parser::statement(Lexer &lexer) {
  switch (lexer.getType()) {
  case TPLID_OBRACE: return blockStmt(lexer);
  case TPLID_IF: return ifStmt(lexer);
  case TPLID_FOR: return forStmt(lexer);
  case TPLID_WHILE: return whileStmt(lexer);
  case TPLID_DO: return doStmt(lexer);
  case TPLID_SWITCH: return switchStmt(lexer);
  case TPLID_BREAK: return breakStmt(lexer);
  case TPLID_CONTINUE: return continueStmt(lexer);
  case TPLID_RETURN: return returnStmt(lexer);
  case TPLID_DOT: return assignStmt(lexer);
  case TPLID_ID: {
    pendingName = name(lexer);

    switch (lexer.getType()) {
    case TPLID_OPAREN: return new ExpressionStmt(callExpr(lexer));
    case TPLID_OBRACKET: case TPLID_ASSIGN: case TPLID_ADD_ASSIGN:
    case TPLID_SUB_ASSIGN: case TPLID_DIV_ASSIGN: case TPLID_MUL_ASSIGN:
    case TPLID_MOD_ASSIGN: return assignStmt(lexer);
    default: break; // Fall through
    }
  }

  default: THROWS("Unexpected token: " << lexer.getType());
  }
}


SmartPointer<BlockStmt> Parser::blockStmt(Lexer &lexer) {
  ParseScope scope(lexer.getScanner());
  SmartPointer<BlockStmt> block = new BlockStmt;

  lexer.match(TPLID_OBRACE);
  blockContents(lexer, *block);
  lexer.match(TPLID_CBRACE);

  return scope.set(block);
}


void Parser::blockContents(Lexer &lexer, BlockStmt &block) {
  while (true) {
    switch (lexer.getType()) {
    case TPLID_EOF: return;
    case TPLID_CBRACE: return;
    case TPLID_FROM: block.push_back(partialImportDecl(lexer)); break;
    case TPLID_IMPORT: block.push_back(importDecl(lexer)); break;
    case TPLID_CONSTANT: block.push_back(constantDecl(lexer)); break;
    case TPLID_FUNCTION: block.push_back(functionDecl(lexer)); break;
    default: block.push_back(statement(lexer)); break;
    }
  }
}


SmartPointer<IfStmt> Parser::ifStmt(Lexer &lexer) {
  ParseScope scope(lexer.getScanner());

  lexer.match(TPLID_IF);
  SmartPointer<Expression> expr = quotedExpr(lexer)->getExpression();
  SmartPointer<Statement> ifStmt = statement(lexer);

  SmartPointer<Statement> elseStmt;
  if (lexer.consume(TPLID_ELSE)) elseStmt = statement(lexer);

  return scope.set(new IfStmt(expr, ifStmt, elseStmt));
}



SmartPointer<ForStmt> Parser::forStmt(Lexer &lexer) {
  ParseScope scope(lexer.getScanner());

  lexer.match(TPLID_FOR);
  lexer.match(TPLID_OPAREN);

  SmartPointer<AssignStmt> init;
  if (lexer.getType() != TPLID_SEMI) init = assignStmt(lexer);

  lexer.match(TPLID_SEMI);

  SmartPointer<Expression> condition;
  if (lexer.getType() != TPLID_SEMI) condition = expression(lexer);

  lexer.match(TPLID_SEMI);

  SmartPointer<AssignStmt> increment;
  if (lexer.getType() != TPLID_SEMI) increment = assignStmt(lexer);

  lexer.match(TPLID_CPAREN);

  return scope.set(new ForStmt(init, condition, increment, statement(lexer)));
}



SmartPointer<WhileStmt> Parser::whileStmt(Lexer &lexer) {
  ParseScope scope(lexer.getScanner());

  lexer.match(TPLID_WHILE);
  SmartPointer<Expression> expr = quotedExpr(lexer)->getExpression();
  SmartPointer<Statement> stmt = statement(lexer);

  return scope.set(new WhileStmt(expr, stmt));
}



SmartPointer<DoStmt> Parser::doStmt(Lexer &lexer) {
  ParseScope scope(lexer.getScanner());

  lexer.match(TPLID_DO);
  SmartPointer<Statement> stmt = statement(lexer);
  lexer.match(TPLID_WHILE);
  SmartPointer<Expression> expr = quotedExpr(lexer)->getExpression();

  return scope.set(new DoStmt(stmt, expr));
}



SmartPointer<SwitchStmt> Parser::switchStmt(Lexer &lexer) {
  ParseScope scope(lexer.getScanner());

  lexer.match(TPLID_SWITCH);
  SmartPointer<SwitchStmt> stmt =
    new SwitchStmt(quotedExpr(lexer)->getExpression());
  lexer.match(TPLID_OBRACE);

  while (lexer.getType() == TPLID_CASE)
    stmt->push_back(caseStmt(lexer));

  if (lexer.consume(TPLID_DEFAULT)) {
    lexer.match(TPLID_COLON);
    stmt->setDefault(statement(lexer));
  }

  lexer.match(TPLID_CBRACE);

  return scope.set(stmt);
}


SmartPointer<CaseStmt> Parser::caseStmt(Lexer &lexer) {
  ParseScope scope(lexer.getScanner());

  lexer.consume(TPLID_CASE);
  CaseStmt::exprs_t exprs;

  while (true) {
    switch (lexer.getType()) {
    case TPLID_STRING: case TPLID_INTEGER:
      exprs.push_back(constant(lexer));
      break;

    case TPLID_ID:
      exprs.push_back(variableExpr(lexer));
      if (!exprs.back()->isConstant())
        THROW("Non-constant variable in case statement");
      break;

    default: THROWS("Expected constant string, integer or variable reference "
                    "in case statement, found " << lexer.getType());
    }

    if (!lexer.consume(TPLID_COMMA)) break;
  }

  lexer.match(TPLID_COLON);

  return scope.set(new CaseStmt(exprs, statement(lexer)));
}


SmartPointer<BreakStmt> Parser::breakStmt(Lexer &lexer) {
  ParseScope scope(lexer.getScanner());
  lexer.match(TPLID_BREAK);
  return scope.set(new BreakStmt);
}


SmartPointer<ContinueStmt> Parser::continueStmt(Lexer &lexer) {
  ParseScope scope(lexer.getScanner());
  lexer.match(TPLID_CONTINUE);
  return scope.set(new ContinueStmt);
}


SmartPointer<ReturnStmt> Parser::returnStmt(Lexer &lexer) {
  ParseScope scope(lexer.getScanner());

  lexer.match(TPLID_RETURN);

  SmartPointer<Expression> expr;
  if (lexer.getType() == TPLID_OPAREN) {
    lexer.advance();
    expr = expression(lexer);
    lexer.match(TPLID_CPAREN);
  }

  return scope.set(new ReturnStmt(expr));
}


SmartPointer<AssignStmt> Parser::assignStmt(Lexer &lexer) {
  ParseScope scope(lexer.getScanner());

  bool global = lexer.consume(TPLID_DOT);

  SmartPointer<Name> nameNode = nextName(lexer);
  scope.setStart(nameNode->getLocation().getStart());

  AssignStmt::indices_t indices;
  while (lexer.consume(TPLID_OBRACKET)) {
    indices.push_back(expression(lexer));
    lexer.match(TPLID_CBRACKET);
  }

  TPLID type = lexer.getType();
  switch (type) {
  case TPLID_ASSIGN:
  case TPLID_ADD_ASSIGN:
  case TPLID_SUB_ASSIGN:
  case TPLID_DIV_ASSIGN:
  case TPLID_MUL_ASSIGN:
  case TPLID_MOD_ASSIGN:
    break;

  default:
    // TODO this could be hit on a non-assignment syntax error
    THROW("Expected assignment operator");
  }
  lexer.advance();

  if (nameNode->getName().size() != 1) THROW("Cannot assign to qualified name");
  string name = nameNode->getName()[0];

  // Declare global variables in parent environment
  Environment &env = global ? getRootEnv() : getEnv();

  SmartPointer<Declaration> decl = env.declareVariable(name);
  if (decl->isConstant()) THROW("Cannot assign to constant");

  SmartPointer<VariableExpr> var = new VariableExpr(nameNode, decl);
  var->getLocation() = nameNode->getLocation();

  return
    scope.set(new AssignStmt(type, var, indices, expression(lexer), global));
}


SmartPointer<Expression> Parser::expression(Lexer &lexer) {
  return boolExpr(lexer);
}


SmartPointer<Expression> Parser::boolExpr(Lexer &lexer) {
  ParseScope scope(lexer.getScanner());
  SmartPointer<Expression> expr = compareExpr(lexer);

  while (lexer.getType() == TPLID_AND || lexer.getType() == TPLID_OR) {
    TPLID type = lexer.getType();

    SmartPointer<Expression> left = expr;
    lexer.advance();
    SmartPointer<Expression> right = compareExpr(lexer);

    expr = new BinaryExpr(type, left, right);
  }

  return scope.set(expr);
}


SmartPointer<Expression> Parser::compareExpr(Lexer &lexer) {
  ParseScope scope(lexer.getScanner());
  SmartPointer<Expression> expr = equalExpr(lexer);

  while (lexer.getType() == TPLID_LESS || lexer.getType() == TPLID_GREATER ||
         lexer.getType() == TPLID_LESS_EQUAL ||
         lexer.getType() == TPLID_GREATER_EQUAL) {
    TPLID type = lexer.getType();

    SmartPointer<Expression> left = expr;
    lexer.advance();
    SmartPointer<Expression> right = equalExpr(lexer);

    expr = new BinaryExpr(type, left, right);
  }

  return scope.set(expr);
}


SmartPointer<Expression> Parser::equalExpr(Lexer &lexer) {
  ParseScope scope(lexer.getScanner());
  SmartPointer<Expression> expr = addExpr(lexer);

  while (lexer.getType() == TPLID_EQUAL || lexer.getType() == TPLID_NOT_EQUAL) {
    TPLID type = lexer.getType();

    SmartPointer<Expression> left = expr;
    lexer.advance();
    SmartPointer<Expression> right = addExpr(lexer);

    expr = new BinaryExpr(type, left, right);
  }

  return scope.set(expr);
}


SmartPointer<Expression> Parser::addExpr(Lexer &lexer) {
  ParseScope scope(lexer.getScanner());
  SmartPointer<Expression> expr = mulExpr(lexer);

  while (lexer.getType() == TPLID_ADD || lexer.getType() == TPLID_SUB) {
    TPLID type = lexer.getType();

    SmartPointer<Expression> left = expr;
    lexer.advance();
    SmartPointer<Expression> right = mulExpr(lexer);

    expr = new BinaryExpr(type, left, right);
  }

  return scope.set(expr);
}


SmartPointer<Expression> Parser::mulExpr(Lexer &lexer) {
  ParseScope scope(lexer.getScanner());
  SmartPointer<Expression> expr = unaryExpr(lexer);

  while (lexer.getType() == TPLID_MUL || lexer.getType() == TPLID_DIV ||
         lexer.getType() == TPLID_MOD) {
    TPLID type = lexer.getType();

    SmartPointer<Expression> left = expr;
    lexer.advance();
    SmartPointer<Expression> right = unaryExpr(lexer);

    expr = new BinaryExpr(type, left, right);
  }

  return scope.set(expr);
}


SmartPointer<Expression> Parser::unaryExpr(Lexer &lexer) {
  if ((lexer.getType() == TPLID_SUB || lexer.getType() == TPLID_NOT) &&
      pendingName.isNull()) {
    TPLID type = lexer.getType();
    ParseScope scope(lexer.getScanner());

    lexer.advance();

    return scope.set(new UnaryExpr(type, unaryExpr(lexer)));
  }

  return sliceExpr(lexer);
}


SmartPointer<Expression> Parser::sliceExpr(Lexer &lexer) {
  SmartPointer<Expression> expr = primary(lexer);

  while (lexer.getType() == TPLID_OBRACKET) {
    ParseScope scope(lexer.getScanner());
    lexer.advance();

    SmartPointer<Expression> start = expression(lexer);

    SmartPointer<Expression> end;
    if (lexer.getType() == TPLID_COLON) {
      lexer.advance();
      end = expression(lexer);
    }

    lexer.match(TPLID_CBRACKET);

    return scope.set(new SliceExpr(expr, start, end));
  }

  return expr;
}


SmartPointer<Expression> Parser::primary(Lexer &lexer) {
  if (!pendingName.isNull()) {
    if (lexer.getType() == TPLID_OPAREN) return callExpr(lexer);
    else return variableExpr(lexer);
  }

  switch (lexer.getType()) {
  case TPLID_OBRACKET: return vectorExpr(lexer);
  case TPLID_OPAREN: return quotedExpr(lexer);
  case TPLID_ID: {
    ParseScope scope(lexer.getScanner());
    pendingName = name(lexer);
    if (lexer.getType() == TPLID_OPAREN) return callExpr(lexer);
    else return variableExpr(lexer);
  }

  default:
    try {
      return constant(lexer);
    } catch (const Exception &e) {
      THROWC("Expected primary", e);
    }
  }
}


SmartPointer<VariableExpr> Parser::variableExpr(Lexer &lexer) {
  SmartPointer<Name> name = nextName(lexer);
  SmartPointer<Declaration> decl = getEnv().lookup(*name);

  if (decl->getType() != TPLID_VARIABLE && decl->getType() != TPLID_ARGUMENT &&
      decl->getType() != TPLID_CONSTANT)
    THROWS("'" << *name << "' is a(n) " << decl->getType()
           << " in this context, expected variable");

  SmartPointer<VariableExpr> var = new VariableExpr(name, decl);
  var->getLocation() = name->getLocation();

  return var;
}


SmartPointer<VectorExpr> Parser::vectorExpr(Lexer &lexer) {
  ParseScope scope(lexer.getScanner());
  SmartPointer<VectorExpr> expr = new VectorExpr;

  lexer.match(TPLID_OBRACKET);

  if (lexer.getType() != TPLID_CBRACKET)
    while (true) {
      expr->push_back(expression(lexer));
      if (!lexer.consume(TPLID_COMMA)) break;
    }

  lexer.match(TPLID_CBRACKET);

  return scope.set(expr);
}


SmartPointer<QuotedExpr> Parser::quotedExpr(Lexer &lexer) {
  ParseScope scope(lexer.getScanner());

  lexer.match(TPLID_OPAREN);
  SmartPointer<QuotedExpr> expr = new QuotedExpr(expression(lexer));
  lexer.match(TPLID_CPAREN);

  return scope.set(expr);
}


SmartPointer<CallExpr> Parser::callExpr(Lexer &lexer) {
  ParseScope scope(lexer.getScanner());

  SmartPointer<Name> funcName = nextName(lexer);
  lexer.match(TPLID_OPAREN);

  // Lookup function
  SmartPointer<Declaration> decl = getEnv().lookup(*funcName);
  if (decl->getType() != TPLID_FUNCTION)
    THROWS("'" << *funcName << "' is not a function in this context");

  FunctionDecl &func = *dynamic_cast<FunctionDecl *>(decl.get());

  // Create argument map
  typedef map<string, bool> args_t;
  args_t args;
  for (unsigned i = 0; i < func.size(); i++) args[func[i]->getName()] = false;

  // Create call
  SmartPointer<CallExpr> call = new CallExpr(funcName, decl);
  scope.setStart(funcName->getLocation().getStart());

  // Parse arguments
  bool positionalOk = true;
  if (lexer.getType() != TPLID_CPAREN)
    while (true) {
      CallExpr::pair_t pair;

      if (lexer.getType() == TPLID_ID) {
        SmartPointer<Name> argName = name(lexer);

        if (lexer.consume(TPLID_ASSIGN)) {
          if (argName->getName().size() != 1)
            THROW("Cannot qualify function call argument name");

          pair.first = argName->getName()[0];
          positionalOk = false;

          // Check that the name is a valid argument for this function
          args_t::iterator it = args.find(pair.first);
          if (it == args.end())
            THROWS("'" << pair.first << "' is not a valid argument name for "
                   "call to " << *funcName << "()");

          if (it->second)
            THROWS("Call to " << *funcName << "() has multiple values for "
                   "argument " << pair.first);

          it->second = true; // Mark used

        } else pendingName = argName;
      }

      // Mark positional argument
      if (pair.first.empty()) {
        if (!positionalOk)
          THROW("Positional argument not allowed after keyword argument");

        if (call->size() < func.size()) {
          const string &name = func[call->size()]->getName();
          args_t::iterator it = args.find(name);
          if (it == args.end()) THROW("Internal error");
          it->second = true; // Make used
        }
      }

      pair.second = expression(lexer);
      call->push_back(pair);

      if (!lexer.consume(TPLID_COMMA)) break;
    }

  lexer.match(TPLID_CPAREN);

  // Check that the number of arguments is reasonable
  if (!func.getVariableArgs() && func.size() < call->size())
    THROWS(func.getName() << "() takes at most " << func.size()
           << " arguments, " << call->size() << " given");

  // Make sure all arguments have a value
  for (unsigned i = 0; i < func.size(); i++) {
    const string &name = func[i]->getName();
    if (!args[name] && func[i]->getDefault().isNull())
      THROWS("Call to " << *funcName << "() did not assign a value to "
             << (name.empty() ? String(i) : name) << " and it has no default");
  }

  return scope.set(call);
}


SmartPointer<Constant> Parser::constant(Lexer &lexer) {
  ParseScope scope(lexer.getScanner());
  SmartPointer<Value> value;
  string text = lexer.getValue();

  switch (lexer.getType()) {
  case TPLID_INTEGER: value = new IntegerValue(String::parseU32(text)); break;
  case TPLID_FLOAT: value = new FloatValue(String::parseDouble(text)); break;
  case TPLID_STRING: value = new StringValue(text); break;
  case TPLID_TRUE: value = new BoolValue(true); break;
  case TPLID_FALSE: value = new BoolValue(false); break;
  case TPLID_NULL: value = NullValue::instance(); break;
  default: THROW("Expected constant expression");
  }
  lexer.advance();

  return scope.set(new Constant(value, text));
}


SmartPointer<Name> Parser::name(Lexer &lexer) {
  ParseScope scope(lexer.getScanner());

  SmartPointer<Name> name = new Name;

  while (true) {
    name->append(lexer.match(TPLID_ID));
    if (!lexer.consume(TPLID_DOT)) break;
  }

  return scope.set(name);
}


cb::SmartPointer<Name> Parser::nextName(Lexer &lexer) {
  if (pendingName.isNull()) return name(lexer);

  SmartPointer<Name> tmp = pendingName;
  pendingName = 0;
  return tmp;
}


SmartPointer<Module> Parser::load(const string name) {
  if (loader.isNull()) THROW("Parser does not have a Loader");

  SmartPointer<Module> module = loader->load(name, cwd);

  if (!module->isLoaded()) {
    module->setLoaded(true);
    Parser parser(envStack.front(), loader);
    EnvScope envScope(parser.envStack, module->getEnvironment());
    cb::SmartPointer<BlockStmt> block = parser.parse(module->getPath());
    module->setBlock(block);
  }

  return module;
}
