#ifndef OPENSCAM_TPL_PARSER_H
#define OPENSCAM_TPL_PARSER_H

#include "Lexer.h"
#include "TPLID.h"
#include "Environment.h"

#include "ast/AST.h"

#include <cbang/SmartPointer.h>
#include <cbang/io/InputSource.h>

#include <map>
#include <string>


namespace OpenSCAM {
  namespace TPL {
    class ArgumentDecls;
    class Module;
    class Loader;
    class Context;

    class Parser : public TPLID {
    protected:
      typedef std::vector<cb::SmartPointer<Environment> > envStack_t;
      envStack_t envStack;

      cb::SmartPointer<Loader> loader;

      std::string cwd;
      cb::SmartPointer<Name> pendingName;

      Environment &getEnv() {return *envStack.back();}
      Environment &getRootEnv() {return *envStack.front();}

      struct EnvScope {
        cb::SmartPointer<Environment> env;
        envStack_t &envStack;

        EnvScope(envStack_t &envStack,
                 const cb::SmartPointer<Environment> &env = new Environment) :
          env(env), envStack(envStack) {
          if (!envStack.empty()) env->setParent(envStack.back().get());
          envStack.push_back(env);
        }

        ~EnvScope() {envStack.pop_back();}

        const cb::SmartPointer<Environment> &getEnvironment() const
        {return env;}
      };

    public:
      Parser(const cb::SmartPointer<Environment> &env,
             const cb::SmartPointer<Loader> &loader);
      Parser(Context &ctx);
      ~Parser();

      cb::SmartPointer<BlockStmt> parse(Lexer &lexer);
      cb::SmartPointer<BlockStmt> parse(const cb::InputSource &source);
      cb::SmartPointer<Program> program(Lexer &lexer);

      // Declarations
      cb::SmartPointer<ImportDecl> partialImportDecl(Lexer &lexer);
      cb::SmartPointer<ImportDecl> importDecl(Lexer &lexer);
      cb::SmartPointer<ConstantDecl> constantDecl(Lexer &lexer);
      cb::SmartPointer<FunctionDecl> functionDecl(Lexer &lexer);
      bool arguments(Lexer &lexer, ArgumentDecls &args);
      cb::SmartPointer<ArgumentDecl> argument(Lexer &lexer);

      // Statements
      cb::SmartPointer<Statement> statement(Lexer &lexer);
      cb::SmartPointer<BlockStmt> blockStmt(Lexer &lexer);
      void blockContents(Lexer &lexer, BlockStmt &block);
      cb::SmartPointer<IfStmt> ifStmt(Lexer &lexer);
      cb::SmartPointer<ForStmt> forStmt(Lexer &lexer);
      cb::SmartPointer<WhileStmt> whileStmt(Lexer &lexer);
      cb::SmartPointer<DoStmt> doStmt(Lexer &lexer);
      cb::SmartPointer<SwitchStmt> switchStmt(Lexer &lexer);
      cb::SmartPointer<CaseStmt> caseStmt(Lexer &lexer);
      cb::SmartPointer<BreakStmt> breakStmt(Lexer &lexer);
      cb::SmartPointer<ContinueStmt> continueStmt(Lexer &lexer);
      cb::SmartPointer<ReturnStmt> returnStmt(Lexer &lexer);
      cb::SmartPointer<AssignStmt> assignStmt(Lexer &lexer);

      // Expressions
      cb::SmartPointer<Expression> expression(Lexer &lexer);
      cb::SmartPointer<Expression> boolExpr(Lexer &lexer);
      cb::SmartPointer<Expression> compareExpr(Lexer &lexer);
      cb::SmartPointer<Expression> equalExpr(Lexer &lexer);
      cb::SmartPointer<Expression> addExpr(Lexer &lexer);
      cb::SmartPointer<Expression> mulExpr(Lexer &lexer);
      cb::SmartPointer<Expression> unaryExpr(Lexer &lexer);

      cb::SmartPointer<Expression> sliceExpr(Lexer &lexer);
      cb::SmartPointer<Expression> primary(Lexer &lexer);
      cb::SmartPointer<VariableExpr> variableExpr(Lexer &lexer);
      cb::SmartPointer<VectorExpr> vectorExpr(Lexer &lexer);
      cb::SmartPointer<QuotedExpr> quotedExpr(Lexer &lexer);
      cb::SmartPointer<CallExpr> callExpr(Lexer &lexer);

      cb::SmartPointer<Constant> constant(Lexer &lexer);
      cb::SmartPointer<Name> name(Lexer &lexer);

    protected:
      cb::SmartPointer<Name> nextName(Lexer &lexer);
      cb::SmartPointer<Module> load(const std::string name);
   };
  }
}

#endif // OPENSCAM_TPL_PARSER_H

