#include "PrettyPrinter.h"

#include "ast/AST.h"

#include <limits>

using namespace OpenSCAM::TPL;
using namespace cb;
using namespace std;


void PrettyPrinter::visit(Program &node) {
  for (unsigned i = 0; i < node.size(); i++) {
    node.at(i)->visit(*this);
    newLine();
  }
}


void PrettyPrinter::visit(ImportDecl &node) {
  if (node.getID().empty()) stream << "import " << node.getName();
  else stream << "from " << node.getName() << " import " << node.getID();

  if (!node.getAs().empty()) stream << " as " << node.getAs();
}


void PrettyPrinter::visit(ConstantDecl &node) {
  stream << "constant " << node.getName() << " = ";
  node.getExpression()->visit(*this);
}


void PrettyPrinter::visit(ArgumentDecl &node) {
  stream << node.getName();
  if (node.getDefault().isNull()) return;
  stream << " = ";
  node.getDefault()->visit(*this);
}


void PrettyPrinter::visit(FunctionDecl &node) {
  stream << "function " << node.getName() << '(';

  for (unsigned i = 0; i < node.size(); i++) {
    if (i) stream << ", ";
    node[i]->visit(*this);
  }

  stream << ") ";
  node.getBlock()->visit(*this);
}


void PrettyPrinter::visit(AssignStmt &node) {
  if (node.isGlobal()) stream << '.';
  stream << *node.getVariable()->getName();

  for (unsigned i = 0; i < node.getIndices().size(); i++) {
    stream << '[';
    node.getIndices()[i]->visit(*this);
    stream << ']';
  }

  stream << ' ';
  switch (node.getType()) {
  case TPLID_ASSIGN: break;
  case TPLID_ADD_ASSIGN: stream << '+'; break;
  case TPLID_SUB_ASSIGN: stream << '-'; break;
  case TPLID_DIV_ASSIGN: stream << '/'; break;
  case TPLID_MUL_ASSIGN: stream << '*'; break;
  case TPLID_MOD_ASSIGN: stream << '%'; break;
  default: THROWS("Invalid assignment statement at " << node.getLocation());
  }
  stream << "= ";

  node.getExpression()->visit(*this);
}


void PrettyPrinter::visit(BlockStmt &node) {
  stream << '{';
  indent++;

  for (unsigned i = 0; i < node.size(); i++) {
    newLine();
    node[i]->visit(*this);
  }

  indent--;
  if (!node.empty()) newLine();
  stream << '}';
}


void PrettyPrinter::visit(BreakStmt &node) {
  stream << "break";
}


void PrettyPrinter::visit(CaseStmt &node) {
  stream << "case ";

  const CaseStmt::exprs_t &exprs = node.getExpressions();
  for (unsigned i = 0; i < exprs.size(); i++) {
    if (i) stream << ", ";
    exprs[i]->visit(*this);
  }

  stream << ": ";
  node.getStatement()->visit(*this);
}


void PrettyPrinter::visit(ContinueStmt &node) {
  stream << "continue";
}


void PrettyPrinter::visit(DoStmt &node) {
  stream << "do ";
  node.getStatement()->visit(*this);
  stream << " while (";
  node.getExpression()->visit(*this);
  stream << ')';
}


void PrettyPrinter::visit(ExpressionStmt &node) {
  node.getExpression()->visit(*this);
}


void PrettyPrinter::visit(ForStmt &node) {
  stream << "for (";

  if (!node.getInit().isNull()) node.getInit()->visit(*this);

  stream << "; ";
  if (!node.getCondition().isNull()) node.getCondition()->visit(*this);

  stream << "; ";
  if (!node.getIncrement().isNull()) node.getIncrement()->visit(*this);

  stream << ") ";
  node.getStatement()->visit(*this);
}


void PrettyPrinter::visit(IfStmt &node) {
  stream << "if (";
  node.getExpression()->visit(*this);
  stream << ") ";
  node.getIfStmt()->visit(*this);

  if (!node.getElseStmt().isNull()) {
    newLine();
    stream << "else ";
    node.getElseStmt()->visit(*this);
  }
}


void PrettyPrinter::visit(ReturnStmt &node) {
  stream << "return";

  if (!node.getExpression().isNull()) {
    stream << " (";
    node.getExpression()->visit(*this);
    stream << ')';
  }
}


void PrettyPrinter::visit(SwitchStmt &node) {
  stream << "switch (";
  node.getExpression()->visit(*this);
  stream << ") {";

  for (unsigned i = 0; i < node.size(); i++) {
    newLine();
    node[i]->visit(*this);
  }

  if (!node.getDefault().isNull()) {
    newLine();
    stream << "default: ";
    node.getDefault()->visit(*this);
  }

  if (!node.empty() || !node.getDefault().isNull()) newLine();

  stream << '}';
}


void PrettyPrinter::visit(WhileStmt &node) {
  stream << "while (";
  node.getExpression()->visit(*this);
  stream << ") ";
  node.getStatement()->visit(*this);
}


void PrettyPrinter::visit(BinaryExpr &node) {
  node.getLeft()->visit(*this);

  stream << ' ';
  switch (node.getType()) {
  case TPLID_AND: stream << "&&"; break;
  case TPLID_OR: stream << "||"; break;
  case TPLID_LESS: stream << '<'; break;
  case TPLID_GREATER: stream << '>'; break;
  case TPLID_LESS_EQUAL: stream << "<="; break;
  case TPLID_GREATER_EQUAL: stream << ">="; break;
  case TPLID_EQUAL: stream << "=="; break;
  case TPLID_NOT_EQUAL: stream << "!="; break;
  case TPLID_ADD: stream << "+"; break;
  case TPLID_SUB: stream << "-"; break;
  case TPLID_MUL: stream << "*"; break;
  case TPLID_DIV: stream << "/"; break;
  case TPLID_MOD: stream << "%"; break;
  default: THROWS("Invalid binary expression at " << node.getLocation());
  }
  stream << ' ';

  node.getRight()->visit(*this);
}


void PrettyPrinter::visit(CallExpr &node) {
  stream << *node.getName() << '(';

  for (unsigned i = 0; i < node.size(); i++) {
    if (i) stream << ", ";
    if (!node[i].first.empty()) stream << node[i].first << " = ";
    node[i].second->visit(*this);
  }

  stream << ')';
}


void PrettyPrinter::visit(QuotedExpr &node) {
  stream << '(';
  node.getExpression()->visit(*this);
  stream << ')';
}


void PrettyPrinter::visit(SliceExpr &node) {
  node.getExpression()->visit(*this);
  stream << '[';
  node.getStart()->visit(*this);

  if (!node.getEnd().isNull()) {
    stream << ':';
    node.getEnd()->visit(*this);
  }

  stream << ']';
}


void PrettyPrinter::visit(UnaryExpr &node) {
  switch (node.getType()) {
  case TPLID_SUB: stream << '-'; break;
  case TPLID_NOT: stream << '!'; break;
  default: THROWS("Invalid unary expression at " << node.getLocation());
  }

  node.getExpression()->visit(*this);
}


void PrettyPrinter::visit(VariableExpr &node) {
  stream << *node.getName();
}


void PrettyPrinter::visit(VectorExpr &node) {
  stream << '[';

  for (unsigned i = 0; i < node.size(); i++) {
    if (i) stream << ", ";
    node[i]->visit(*this);
  }

  stream << ']';
}


void PrettyPrinter::visit(Constant &node) {
  if (node.getValue()->isString())
    stream << '"' << String::escapeC(node.getText()) << '"';
  else stream << node.getText();
}


void PrettyPrinter::newLine() {
  stream << '\n';
  for (unsigned i = 0; i < indent; i++) stream << "  ";
}
