#include "DXFLibrary.h"

#include <openscam/tpl/type/VectorValue.h>

#include <cbang/log/Logger.h>
#include <cbang/geom/Vector.h>

#include <dxflib/dl_dxf.h>
#include <dxflib/dl_creationadapter.h>

#include <set>

#include <math.h>

using namespace std;
using namespace cb;
using namespace OpenSCAM::TPL;


namespace {
  enum {
    DXF_POINT,
    DXF_LINE,
    DXF_ARC,
    DXF_POLYLINE,
  };

  class DXFReader : public DL_CreationAdapter {
    typedef map<string, unsigned> layerMap_t;
    layerMap_t layerMap;
    SmartPointer<VectorValue> layers;
    SmartPointer<VectorValue> polyline;

    Vector3D basePt;

  public:
    DXFReader() : layers(new VectorValue) {}

    const SmartPointer<VectorValue> &getLayers() const {return layers;}


    SmartPointer<VectorValue> addEntity(unsigned type) {
      SmartPointer<VectorValue> entity = new VectorValue;
      entity->appendInteger(type);

      layerMap_t::iterator it = layerMap.find(attributes.getLayer());
      if (it == layerMap.end())
        THROWS("DXF Undefined layer '" << attributes.getLayer() << "'");

      layers->at(it->second)->toVector().at(1)->append(entity);

      return entity;
    }


    static SmartPointer<VectorValue> makeVertex(double x, double y, double z) {
      SmartPointer<VectorValue> v = new VectorValue;

      v->appendFloat(x);
      v->appendFloat(y);
      v->appendFloat(z);

      return v;
    }


    // From DL_CreationAdapter
    void addLayer(const DL_LayerData &data) {
      layerMap[data.name] = layers->size();

      SmartPointer<VectorValue> layer = new VectorValue;
      layer->appendString(data.name);
      layer->append(new VectorValue);
      layers->append(layer);
    }


    void addPoint(const DL_PointData &point) {
      addEntity(DXF_POINT)->append(makeVertex(point.x, point.y, point.z));
    }


    void addLine(const DL_LineData &line) {
      SmartPointer<VectorValue> l = new VectorValue;
      l->append(makeVertex(line.x1, line.y1, line.z1));
      l->append(makeVertex(line.x2, line.y2, line.z2));
      addEntity(DXF_LINE)->append(l);
    }


    void addArc(const DL_ArcData &arc) {
      SmartPointer<VectorValue> a = new VectorValue;
      a->append(makeVertex(arc.cx, arc.cy, arc.cz));
      a->appendFloat(arc.radius);
      a->appendFloat(arc.angle1);
      a->appendFloat(arc.angle2);
      addEntity(DXF_ARC)->append(a);
    }


    void addCircle(const DL_CircleData &circle) {
      SmartPointer<VectorValue> a = new VectorValue;
      a->append(makeVertex(circle.cx, circle.cy, circle.cz));
      a->appendFloat(circle.radius);
      a->appendFloat(0);
      a->appendFloat(2 * M_PI);
      addEntity(DXF_ARC)->append(a);
    }


    void addPolyline(const DL_PolylineData &polyline) {
      if (!this->polyline.isNull()) THROW("DXF Already in polyline");
      this->polyline = new VectorValue;
      addEntity(DXF_POLYLINE)->append(this->polyline);
    }


    void addVertex(const DL_VertexData &vertex) {
      if (polyline.isNull()) return;
      polyline->append(makeVertex(vertex.z, vertex.y, vertex.z));
    }


    void endEntity() {polyline = 0;}


    void addEllipse(const DL_EllipseData &ellipse) {
      LOG_WARNING("DXF Ellipse not supported");
    }


    void add3dFace(const DL_3dFaceData &face) {
      LOG_WARNING("DXF 3D Face not supported");
    }
  };
}


DXFLibrary::DXFLibrary() : Module("dxf") {
#define DECLARE_FUNC(name, args)                    \
  OPENSCAM_TPL_DECLARE_FUNC(DXFLibrary, name, args)

  DECLARE_FUNC(open, "path");

  Environment &env = *getEnvironment();
  env.declareIntegerConstant("POINT",    DXF_POINT);
  env.declareIntegerConstant("LINE",     DXF_LINE);
  env.declareIntegerConstant("ARC",      DXF_ARC);
  env.declareIntegerConstant("POLYLINE", DXF_POLYLINE);
}


SmartPointer<Value> DXFLibrary::open(Context &ctx, const Arguments &args) {
  const string &path = args[0]->toString();
  DXFReader reader;
  DL_Dxf dxf;

  if (!dxf.in(path, &reader)) THROWS("Failed to read '" << path << "' as DXF");

  return reader.getLayers();
}
