#include "GCodeLibrary.h"

#include "../Context.h"
#include "../Environment.h"

#include <openscam/cutsim/Move.h>
#include <openscam/sim/Controller.h>
#include <openscam/sim/Machine.h>

#include <cbang/log/Logger.h>

using namespace std;
using namespace cb;
using namespace OpenSCAM;
using namespace OpenSCAM::TPL;


GCodeLibrary::GCodeLibrary() : Module("gcode") {
#define DECLARE_FUNC(name, args) \
  OPENSCAM_TPL_DECLARE_FUNC(GCodeLibrary, name, args)

  string axes =
    "x=null, y=null, z=null, a=null, b=null, c=null, u=null, v=null, w=null";
  string offsets = "i=null, j=null, k=null";

  DECLARE_FUNC(imperial, "");
  DECLARE_FUNC(metric, "");
  DECLARE_FUNC(rapid, axes);
  DECLARE_FUNC(cut, axes);
  DECLARE_FUNC(arc, axes + "," + offsets + ", clockwise=true");
  DECLARE_FUNC(dwell, "seconds");
  DECLARE_FUNC(plane, "plane=null");
  DECLARE_FUNC(tool, "number");
  DECLARE_FUNC(feed, "rate");
  DECLARE_FUNC(speed, "rate");
  DECLARE_FUNC(spindle, "enable, clockwise=true");
  DECLARE_FUNC(coolant, "mist=null, flood=null");
  DECLARE_FUNC(gcode, "path");

  Environment &env = *getEnvironment();
  env.declareIntegerConstant("XY_PLANE", Controller::XY_PLANE);
  env.declareIntegerConstant("ZX_PLANE", Controller::ZX_PLANE);
  env.declareIntegerConstant("YZ_PLANE", Controller::YZ_PLANE);
  env.declareIntegerConstant("UV_PLANE", Controller::UV_PLANE);
  env.declareIntegerConstant("WU_PLANE", Controller::WU_PLANE);
  env.declareIntegerConstant("VW_PLANE", Controller::VW_PLANE);
}


SmartPointer<Value> GCodeLibrary::imperial(Context &ctx,
                                           const Arguments &args) {
  ctx.getController().setMetric(false);
  return 0;
}


SmartPointer<Value> GCodeLibrary::metric(Context &ctx, const Arguments &args) {
  ctx.getController().setMetric(true);
  return 0;
}


static int setVars(Context &ctx, const Arguments &args, const char *letters) {
  Controller &controller = ctx.getController();
  int vars = 0;

  for (unsigned i = 0; letters[i] && i < args.size(); i++)
    if (letters[i] != ' ' && !args[i]->isNull()) {
      vars |= Controller::letterToVarType(letters[i]);
      controller.setVar(letters[i], args[i]->toFloat());
    }

  return vars;
}


static void makeMove(MoveType type, Context &ctx, const Arguments &args) {
  Controller &controller = ctx.getController();
  controller.makeMove(type, setVars(ctx, args, "XYZABCUVW"));
}


SmartPointer<Value> GCodeLibrary::rapid(Context &ctx, const Arguments &args) {
  makeMove(MoveType::MOVE_RAPID, ctx, args);
  return 0;
}


SmartPointer<Value> GCodeLibrary::cut(Context &ctx, const Arguments &args) {
  makeMove(MoveType::MOVE_CUTTING, ctx, args);
  return 0;
}


SmartPointer<Value> GCodeLibrary::arc(Context &ctx, const Arguments &args) {
  Controller &controller = ctx.getController();

  int vars = setVars(ctx, args, "XYZABCUVWIJK");
  bool clockwise = args[12]->toBool();

  controller.arc(vars, clockwise);

  return 0;
}


SmartPointer<Value> GCodeLibrary::dwell(Context &ctx, const Arguments &args) {
  ctx.getController().dwell(args[0]->toFloat());
  return 0;
}


SmartPointer<Value> GCodeLibrary::plane(Context &ctx, const Arguments &args) {
  if (!args[0]->isNull())
    ctx.getController().setPlane((Controller::plane_t)args[0]->toInteger());
  return new IntegerValue(ctx.getController().getPlane());
}


SmartPointer<Value> GCodeLibrary::tool(Context &ctx, const Arguments &args) {
  ctx.getController().setTool((unsigned)args[0]->toInteger());
  return 0;
}


SmartPointer<Value> GCodeLibrary::feed(Context &ctx, const Arguments &args) {
  ctx.getController().setFeed(args[0]->toFloat());
  return 0;
}


SmartPointer<Value> GCodeLibrary::speed(Context &ctx, const Arguments &args) {
  ctx.getController().setSpeed(args[0]->toFloat());
  return 0;
}


SmartPointer<Value> GCodeLibrary::spindle(Context &ctx, const Arguments &args) {
  Controller &controller = ctx.getController();
  bool enable = args[0]->toBool();
  bool clockwise = args[1]->toBool();

  if (enable) controller.setSpindleDir(clockwise ? Controller::DIR_CLOCKWISE :
                                       Controller::DIR_COUNTERCLOCKWISE);
  else controller.setSpindleDir(Controller::DIR_OFF);

  return 0;
}


SmartPointer<Value> GCodeLibrary::coolant(Context &ctx, const Arguments &args) {
  if (!args[0]->isNull())
    ctx.getController().setMistCoolant(args[0]->toBool());

  if (!args[1]->isNull())
    ctx.getController().setFloodCoolant(args[1]->toBool());

  return 0;
}


SmartPointer<Value> GCodeLibrary::gcode(Context &ctx, const Arguments &args) {
  return 0;
}
