#include "STDLibrary.h"

#include "../Context.h"
#include "../Interpreter.h"

using namespace std;
using namespace cb;
using namespace OpenSCAM::TPL;


STDLibrary::STDLibrary() : Module("std") {
#define DECLARE_FUNC(name, args)                    \
  OPENSCAM_TPL_DECLARE_FUNC(STDLibrary, name, args)

  DECLARE_FUNC(size,       "x");

  DECLARE_FUNC(to_bool,    "x");
  DECLARE_FUNC(to_integer, "x");
  DECLARE_FUNC(to_float,   "x");
  DECLARE_FUNC(to_string,  "x");

  DECLARE_FUNC(is_null,    "x");
  DECLARE_FUNC(is_bool,    "x");
  DECLARE_FUNC(is_number,  "x");
  DECLARE_FUNC(is_integer, "x");
  DECLARE_FUNC(is_float,   "x");
  DECLARE_FUNC(is_string,  "x");
  DECLARE_FUNC(is_vector,  "x");

  DECLARE_FUNC(printf, "format, ...");
}


SmartPointer<Value> STDLibrary::size(Context &ctx, const Arguments &args) {
  const Value &value = *args[0];

  switch (value.getType()) {
  case TPLID_STRING: return new IntegerValue(value.toString().length());
  case TPLID_VECTOR: return new IntegerValue(value.toVector().size());
  default: THROWS("Invalid parameter type " << value.getType());
  }
}


SmartPointer<Value> STDLibrary::to_bool(Context &ctx, const Arguments &args) {
  return new BoolValue(args[0]->toBool());
}


SmartPointer<Value> STDLibrary::to_integer(Context &ctx,
                                           const Arguments &args) {
  const Value &value = *args[0];
  int32_t x;

  switch (value.getType()) {
  case TPLID::TPLID_BOOL: x = value.toBool(); break;
  case TPLID::TPLID_STRING: x = String::parseU32(value.toString()); break;
  default: x = value.toInteger(); break;
  }

  return new IntegerValue(x);
}


SmartPointer<Value> STDLibrary::to_float(Context &ctx, const Arguments &args) {
  const Value &value = *args[0];
  double x;

  switch (value.getType()) {
  case TPLID::TPLID_BOOL: x = value.toBool(); break;
  case TPLID::TPLID_STRING: x = String::parseDouble(value.toString()); break;
  default: x = value.toFloat(); break;
  }

  return new FloatValue(x);
}


static string valueToString(const Value &value) {
  switch (value.getType()) {
  case TPLID::TPLID_NULL: return "(null)";
  case TPLID::TPLID_BOOL: return value.toBool() ? "true" : "false";
  case TPLID::TPLID_INTEGER: return String(value.toInteger());
  case TPLID::TPLID_FLOAT: return String(value.toFloat());

  case TPLID::TPLID_VECTOR: {
    string s = "[";

    for (unsigned i = 0; i < value.toVector().size(); i++) {
      const Value &child = *value.toVector()[i];

      if (i) s += ", ";

      if (child.isString()) s += '"';
      s += valueToString(child);
      if (child.isString()) s += '"';
    }

    return s + "]";
  }

  default: return value.toString();
  }
}


SmartPointer<Value> STDLibrary::to_string(Context &ctx, const Arguments &args) {
  return new StringValue(valueToString(*args[0]));
}


SmartPointer<Value> STDLibrary::is_null(Context &ctx, const Arguments &args) {
  return new BoolValue(args[0]->isNull());
}


SmartPointer<Value> STDLibrary::is_bool(Context &ctx, const Arguments &args) {
  return new BoolValue(args[0]->isBool());
}


SmartPointer<Value> STDLibrary::is_number(Context &ctx, const Arguments &args) {
  return new BoolValue(args[0]->isNumber());
}


SmartPointer<Value> STDLibrary::is_integer(Context &ctx,
                                           const Arguments &args) {
  return new BoolValue(args[0]->isInteger());
}


SmartPointer<Value> STDLibrary::is_float(Context &ctx, const Arguments &args) {
  return new BoolValue(args[0]->isFloat());
}


SmartPointer<Value> STDLibrary::is_string(Context &ctx, const Arguments &args) {
  return new BoolValue(args[0]->isString());
}


SmartPointer<Value> STDLibrary::is_vector(Context &ctx, const Arguments &args) {
  return new BoolValue(args[0]->isVector());
}


SmartPointer<Value> STDLibrary::printf(Context &ctx, const Arguments &args) {
  ostream &stream = ctx.getOStream();
  streampos startPos = stream.tellp();

  const string &format = args[0]->toString();
  string::const_iterator it = format.begin();
  unsigned i = 1;

  while (it != format.end())
    if (*it == '%') {
      string::const_iterator start = it++;

      if (*it == '%') {
        stream << '%';
        continue;
      }

      // flag characters
      do {
        switch (*it) {
        case '#': case '0': case '-': case ' ': case '+': it++; continue;
        default: break;
        }
      } while (false);

      // width
      if (*it != '0') while ('0' <= *it && *it <= '9') it++;

      // precision
      if (*it == '.') {
        it++;
        while ('0' <= *it && *it <= '9') it++;
      }

      // conversion
      char c = *it++;
      string conversion(start, it);
      switch (c) {
      case 'b':
        if (args.size() <= i) THROW("Not enough arguments for format");
        stream << (args[i++]->toBool() ? "true" : "false");
        break;

      case 'd': case 'i':
        if (args.size() <= i) THROW("Not enough arguments for format");
        if (!args[i]->isInteger())
          THROWS("printf() '" << c << "' conversion requires integer");
        stream << String::printf(conversion.c_str(), args[i++]->toInteger());
        break;

      case 'o': case 'u': case 'x': case 'X':
        if (args.size() <= i) THROW("Not enough arguments for format");
        if (!args[i]->isInteger())
          THROWS("printf() '" << c << "' conversion requires integer");
        stream << String::printf(conversion.c_str(),
                                 (uint32_t)args[i++]->toInteger());
        break;

      case 'e': case 'E': case 'f': case 'F': case 'g':
        if (args.size() <= i) THROW("Not enough arguments for format");
        if (!args[i]->isFloat())
          THROWS("printf() '" << c << "' conversion requires float");
        stream << String::printf(conversion.c_str(), args[i++]->toFloat());
        break;

      case 's':
        if (args.size() <= i) THROW("Not enough arguments for format");
        if (!args[i]->isString())
          THROW("printf() 's' conversion requires string");
        stream << String::printf(conversion.c_str(),
                                 args[i++]->toString().c_str());
        break;

      default:
        THROWS("Invalid or unsupported printf() conversion: " << conversion);
      }

    } else stream << *it++;

  return new IntegerValue(stream.tellp() - startPos);
}
