/******************************************************************************\

    OpenSCAM is an Open-Source CAM software.
    Copyright (C) 2011-2012 Joseph Coffland <joseph@cauldrondevelopment.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

\******************************************************************************/

#include "Tran.h"

#include <openscam/gcode/Parser.h>

#include <openscam/gcode/ast/Block.h>
#include <openscam/gcode/ast/Word.h>
#include <openscam/gcode/ast/BinaryOp.h>
#include <openscam/gcode/ast/Number.h>
#include <openscam/gcode/ast/QuotedExpr.h>

#include <cbang/String.h>

#include <cbang/config/Options.h>
#include <cbang/log/Logger.h>

#include <cctype>

using namespace std;
using namespace cb;
using namespace OpenSCAM;


Tran::Tran(Options &options, ostream &stream) :
  Printer(stream), options(options) {

  axisSwap[0] = axisSwap[1] = axisSwap[2] = false;

  options.pushCategory("Translate");
  options.add("tran", "Translate axes: tran <X>[,<Y>[,<Z>]]");
  options.add("scale", "Scale axes: scale <X>[,<Y>[,<Z>]]");
  options.addTarget("x-flip", flipOffset[0], "Flip X axis: x-flip <offset>");
  options.addTarget("y-flip", flipOffset[1], "Flip Y axis: y-flip <offset>");
  options.addTarget("z-flip", flipOffset[2], "Flip Z axis: z-flip <offset>");
  options.addTarget("xy-swap", axisSwap[0], "Swap X and Y axes");
  options.addTarget("xz-swap", axisSwap[1], "Swap X and Z axes");
  options.addTarget("yz-swap", axisSwap[2], "Swap Y and Z axes");
  options.popCategory();
}


void Tran::init() {
  // Translate
  unsigned i = 0;
  if (options["tran"].hasValue()) {
    vector<string> parts;
    String::tokenize(options["tran"], parts, " \n\r\t,");
    
    for (; i < parts.size(); i++)
      tran.setIndex(i, String::parseDouble(parts[i]));    
  }
  for (; i < 9; i++) tran.setIndex(i, 0);

  // Scale
  i = 0;
  if (options["scale"].hasValue()) {
    vector<string> parts;
    String::tokenize(options["scale"], parts, " \n\r\t,");

    for (; i < parts.size(); i++)
      scale.setIndex(i, String::parseDouble(parts[i]));
  }
  for (; i < 9; i++) scale.setIndex(i, 1);

  // Flip
  flip[0] = options["x-flip"].isSet();
  flip[1] = options["y-flip"].isSet();
  flip[2] = options["z-flip"].isSet();
}


void Tran::read(const InputSource &source) {
  Parser().parse(source, *this);
}


void Tran::tranWord(Word *word, unsigned axis) {
  if (tran.getIndex(axis)) {
    SmartPointer<Entity> left = new Number(tran.getIndex(axis));
    SmartPointer<Entity> right = word->getExpression();
    SmartPointer<Entity> op = new BinaryOp(Operator::ADD_OP, left, right);
    SmartPointer<Entity> expr = new QuotedExpr(op);
    word->setExpression(eval.simplify(expr));
  }
}


void Tran::scaleWord(Word *word, unsigned axis) {
  if (scale.getIndex(axis) != 1) {
    SmartPointer<Entity> left = new Number(scale.getIndex(axis));
    SmartPointer<Entity> right = word->getExpression();
    SmartPointer<Entity> op = new BinaryOp(Operator::MUL_OP, left, right);
    SmartPointer<Entity> expr = new QuotedExpr(op);
    word->setExpression(eval.simplify(expr));
  }
}


void Tran::flipWord(Word *word, double offset) {
  SmartPointer<Entity> left = new Number(2 * offset);
  SmartPointer<Entity> right = word->getExpression();
  SmartPointer<Entity> op = new BinaryOp(Operator::SUB_OP, left, right);
  SmartPointer<Entity> expr = new QuotedExpr(op);
  word->setExpression(eval.simplify(expr));
}


void Tran::operator()(const SmartPointer<Block> &block) {
  for (unsigned i = 0; i < 9; i++) {
    Word *word = block->findWord(Axes::toAxis(i));
    if (word) {
      tranWord(word, i);
      scaleWord(word, i);
    }
  }

  // TODO is this always correct?  Check for G3?
  // Arc offsets
  const char *off = "IJK";
  const char *axis = "XYZ";
  for (unsigned i = 0; off[i]; i++) {
    Word *word = block->findWord(off[i]);
    if (word) scaleWord(word, Axes::toIndex(axis[i]));
  }

  // Flip
  for (unsigned i = 0; i < 3; i++)
    if (flip[i]) {
      Word *word = block->findWord(axis[i]);
      if (word) flipWord(word, flipOffset[i]);

      // Flip arcs
      // TODO Arcs not correct for Z axis flip!
      word = block->findWord('G');
      if (word) {
        unsigned value = (unsigned)word->eval(eval);
        if (value == 2 || value == 3) {
          // Change direction
          value = value == 3 ? 2 : 3;
          word->setExpression(new Number(value));

          // Flip offset
          word = block->findWord(off[i]);
          if (word) flipWord(word, 0);
        }
      }
    }

  // Swap axes
  for (unsigned i = 0; i < 3; i++)
    if (axisSwap[i]) {
      const char *axis;
      const char *off;

      switch (i) {
      case 0: axis = "XY"; off = "IJ"; break;
      case 1: axis = "XZ"; off = "IK"; break;
      case 2: axis = "YZ"; off = "JK"; break;
      }

      Word *a = block->findWord(axis[0]);
      Word *b = block->findWord(axis[1]);
      if (a) a->setType(axis[1]);
      if (b) b->setType(axis[0]);

      Word *c = block->findWord(off[0]);
      Word *d = block->findWord(off[1]);
      if (c) c->setType(off[1]);
      if (d) d->setType(off[0]);

      Word *word = block->findWord('G');
      if (word) {
        unsigned value = (unsigned)word->eval(eval);
        if (value == 2 || value == 3) {
          // Change direction
          value = value == 3 ? 2 : 3;
          word->setExpression(new Number(value));
        }
      }
    }


  Printer::operator()(block);
}
