/******************************************************************************\

    OpenSCAM is an Open-Source CAM software.
    Copyright (C) 2011-2012 Joseph Coffland <joseph@cauldrondevelopment.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

\******************************************************************************/

#include "Viewer.h"

#include "View.h"
#include "GL.h"
#include "GLFreeType.h"

#include <openscam/sim/ToolTable.h>

#include <cbang/String.h>
#include <cbang/Math.h>
#include <cbang/util/DefaultCatch.h>
#include <cbang/util/HumanNumber.h>
#include <cbang/time/HumanTime.h>

#include <iostream>
#include <fstream>

using namespace std;
using namespace cb;
using namespace OpenSCAM;


static void drawCylinder(GLUquadric *quad, double radiusA, double radiusB,
                         double length) {
  // Body
  gluCylinder(quad, radiusA, radiusB, length, 100, 100);

  // End caps
  if (radiusA) gluDisk(quad, 0, radiusA, 100, 100);
  if (radiusB) {
    glPushMatrix();
    glTranslatef(0, 0, length);
    gluDisk(quad, 0, radiusB, 100, 100);
    glPopMatrix();
  }
}


void Viewer::init() {
  if (!font)
    try {
      font = new GLFreeType("Courier_New.ttf", 12);
    } CBANG_CATCH_ERROR;

  if (!toolQuad) {
    toolQuad = gluNewQuadric();
    if (!toolQuad) THROW("Failed to allocate tool quad");
  }
}


void Viewer::draw(const View &view) {
  init();

  // Setup view port
  view.glDraw(view.path->getBounds());

  // Enable Lighting
  view.setLighting(true);

  // Tool path
  if (view.isFlagSet(View::SHOW_PATH_FLAG)) view.path->draw();
  else view.path->update();

  double currentTime = view.path->getTime();
  real currentDistance = view.path->getDistance();
  Vector3R currentPosition = view.path->getPosition();
  const Move &move = view.path->getMove();

  // Model
  static const float ambientTS[] = {12.0 / 255, 45.0 / 255,  83.0 / 255, 1.00};
  static const float diffuseTS[] = {16.0 / 255, 59.0 / 255, 108.0 / 255, 1.00};

  SmartPointer<Surface> surface = view.renderer->getSurface();
  if (!surface.isNull()) {
    if (view.isFlagSet(View::SHOW_SURFACE_FLAG)) {
      view.setWire(view.isFlagSet(View::WIRE_FLAG));

      glDisable(GL_COLOR_MATERIAL);
      glMaterialfv(GL_FRONT, GL_AMBIENT, ambientTS);
      glMaterialfv(GL_FRONT, GL_DIFFUSE, diffuseTS);

      surface->draw();
      glEnable(GL_COLOR_MATERIAL);

      view.setWire(false);
    }

    if (view.isFlagSet(View::SHOW_NORMALS_FLAG)) {
      glColor4f(1.0, 1.0, 0, 0.6);
      surface->drawNormals();
    }
  }
  if (view.isFlagSet(View::SHOW_BBTREE_FLAG)) view.renderer->drawBB();

  // Tool
  if (view.isFlagSet(View::SHOW_TOOL_FLAG) && !view.path->getPath().isNull()) {
    glTranslatef(currentPosition.x(), currentPosition.y(),
                 currentPosition.z());

    const ToolTable &tools = view.path->getPath()->getTools();
    const Tool &tool = *tools.get(move.getTool());
    double diameter = tool.getDiameter();
    double radius = tool.getRadius();
    double length = tool.getLength();
    ToolShape shape = tool.getShape();

    if (radius <= 0) {
      // Default tool specs
      radius = 25.4 / 8;
      shape = ToolShape::TS_CONICAL;

      glColor4f(1, 0, 0, 1); // Red

    } else glColor4f(1, 0.5, 0, 1); // Orange

    if (length <= 0) length = 50;

    switch (shape) {
    case ToolShape::TS_SPHEROID: {
      glMatrixMode(GL_MODELVIEW);
      glPushMatrix();
      glTranslatef(0, 0, length / 2);
      glScaled(1, 1, length / diameter);
      gluSphere((GLUquadric *)toolQuad, radius, 100, 100);
      glPopMatrix();
      break;
    }

    case ToolShape::TS_BALLNOSE:
      glPushMatrix();
      glTranslatef(0, 0, radius);
      gluSphere((GLUquadric *)toolQuad, radius, 100, 100);
      drawCylinder((GLUquadric *)toolQuad, radius, radius, length - radius);
      glPopMatrix();
      break;

    case ToolShape::TS_SNUBNOSE:
      drawCylinder((GLUquadric *)toolQuad, tool.getSnubDiameter() / 2, radius,
                   length);
      break;

    case ToolShape::TS_CONICAL:
      drawCylinder((GLUquadric *)toolQuad, 0, radius, length);
      break;

    case ToolShape::TS_CYLINDRICAL:
    default:
      drawCylinder((GLUquadric *)toolQuad, radius, radius, length);
      break;
    }
  }

  // Disable Lighting
  view.setLighting(false);

  // Info
  if (view.isFlagSet(View::SHOW_INFO_FLAG) && font) {
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(0, view.getWidth(), 0, view.getHeight());
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glFrontFace(GL_CCW);
    glColor3f(1, 1, 1);

    string direction;
    if (move.getSpeed())
      direction = move.getSpeed() < 0 ? "Counterclockwise" : "Clockwise";

    double totalTime = view.path->getTotalTime();
    real totalDistance = view.path->getTotalDistance();
    double percentDone = currentTime / totalTime * 100;
    uint64_t samples = view.renderer->getSamples();
    uint64_t primitives = surface.isNull() ? 0 : surface->getCount();
    real resolution = view.renderer->getResolution();
    string mode;

    switch (view.renderer->getMode()) {
    case RenderMode::MCUBES_MODE: mode = "Marching Cubes"; break;
    case RenderMode::MTETRA_MODE: mode = "Marching Tetrahedron"; break;
    case RenderMode::DC_MODE: mode = "Dual Contouring"; break;
    case RenderMode::HERMITE_MODE: mode = "Hermite"; break;
    default: break;
    }

    bool play = view.isFlagSet(View::PLAY_FLAG);
    bool loop = view.isFlagSet(View::LOOP_FLAG);

    const unsigned w = 12;
    string info =
      SSTR(""
           << setw(w) << "FPS: " << view.fps << '\n'
           << setw(w) << "Time: " << HumanTime(currentTime) << '/'
           << HumanTime(view.path->getTotalTime()) << '\n'
           << setw(w) << "ETA: " << HumanTime(totalTime - currentTime)
           << '\n'
           << setw(w) << "% Done: " << String::printf("%.02f%%", percentDone)
           << '\n'
           << setw(w) << "Distance: "
           << String::printf("%.04f/%.04fmm\n", currentDistance, totalDistance)
           << setw(w) << "Position: "
           << String::printf("(% 4.04f, % 4.04f, % 4.04f)mm\n",
                             currentPosition.x(),
                             currentPosition.y(),
                             currentPosition.z())
           << setw(w) << "Tool: " << move.getTool() << '\n'
           << setw(w) << "Feed: " << move.getFeed() << "mm/min\n"
           << setw(w) << "Speed: " << fabs(move.getSpeed()) << "RPM\n"
           << setw(w) << "Direction: " << direction << '\n'
           << setw(w) << (play ? "Play: " : "Pause: ") << view.speed
           << 'x' << (loop ? " <->" : "") << '\n'
           << setw(w) << "Resolution: " << resolution << '\n'
           << setw(w) << "Samples: " << HumanNumber(samples) << '\n'
           << setw(w) << "Primitives: " << HumanNumber(primitives) << '\n'
           << setw(w) << "Mode: " << mode << '\n'
           );

#if 0 // TODO fix this
    if (!view.connectionStr.empty()) {
      info +=
        SSTR(setw(w) << "Connected: "
             << (view.connected ? view.connectionStr : string("false"))
             << '\n');

      if (view.connected) {
        info +=
          SSTR(""
               << setw(w) << "Line: " << programLine << '\n'
               << setw(w) << "Status: " << programStatus << '\n'
               << setw(w) << "EStop: " << (estopFlag ? "ON" : "OFF") << '\n'
               << setw(w) << "Machine: " << (machineFlag ? "ON": "OFF") << '\n'
               << setw(w) << "Mode: " << mode << '\n'
               );
      }
    }
#endif

    font->print(0, view.getHeight() - 12, info);
  }

  CHECK_GL_ERROR("");
}
